class PerformanceCounter(Component,IComponent,IDisposable,ISupportInitialize):
 """
 Represents a Windows NT performance counter component.

 

 PerformanceCounter()

 PerformanceCounter(categoryName: str,counterName: str,instanceName: str,machineName: str)

 PerformanceCounter(categoryName: str,counterName: str,instanceName: str)

 PerformanceCounter(categoryName: str,counterName: str,instanceName: str,readOnly: bool)

 PerformanceCounter(categoryName: str,counterName: str)

 PerformanceCounter(categoryName: str,counterName: str,readOnly: bool)
 """
 def BeginInit(self):
  """
  BeginInit(self: PerformanceCounter)

   Begins the initialization of a System.Diagnostics.PerformanceCounter instance used on a form or 

    by another component. The initialization occurs at runtime.
  """
  pass
 def Close(self):
  """
  Close(self: PerformanceCounter)

   Closes the performance counter and frees all the resources allocated by this performance counter 

    instance.
  """
  pass
 @staticmethod
 def CloseSharedResources():
  """
  CloseSharedResources()

   Frees the performance counter library shared state allocated by the counters.
  """
  pass
 def Decrement(self):
  """
  Decrement(self: PerformanceCounter) -> Int64

  

   Decrements the associated performance counter by one through an efficient atomic operation.

   Returns: The decremented counter value.
  """
  pass
 def Dispose(self):
  """ Dispose(self: PerformanceCounter,disposing: bool) """
  pass
 def EndInit(self):
  """
  EndInit(self: PerformanceCounter)

   Ends the initialization of a System.Diagnostics.PerformanceCounter instance that is used on a 

    form or by another component. The initialization occurs at runtime.
  """
  pass
 def GetService(self,*args):
  """
  GetService(self: Component,service: Type) -> object

  

   Returns an object that represents a service provided by the System.ComponentModel.Component or 

    by its System.ComponentModel.Container.

  

  

   service: A service provided by the System.ComponentModel.Component.

   Returns: An System.Object that represents a service provided by the System.ComponentModel.Component,or 

    null if the System.ComponentModel.Component does not provide the specified service.
  """
  pass
 def Increment(self):
  """
  Increment(self: PerformanceCounter) -> Int64

  

   Increments the associated performance counter by one through an efficient atomic operation.

   Returns: The incremented counter value.
  """
  pass
 def IncrementBy(self,value):
  """
  IncrementBy(self: PerformanceCounter,value: Int64) -> Int64

  

   Increments or decrements the value of the associated performance counter by a specified amount 

    through an efficient atomic operation.

  

  

   value: The value to increment by. (A negative value decrements the counter.)

   Returns: The new counter value.
  """
  pass
 def MemberwiseClone(self,*args):
  """
  MemberwiseClone(self: MarshalByRefObject,cloneIdentity: bool) -> MarshalByRefObject

  

   Creates a shallow copy of the current System.MarshalByRefObject object.

  

   cloneIdentity: false to delete the current System.MarshalByRefObject object's identity,which will cause the 

    object to be assigned a new identity when it is marshaled across a remoting boundary. A value of 

    false is usually appropriate. true to copy the current System.MarshalByRefObject object's 

    identity to its clone,which will cause remoting client calls to be routed to the remote server 

    object.

  

   Returns: A shallow copy of the current System.MarshalByRefObject object.

  MemberwiseClone(self: object) -> object

  

   Creates a shallow copy of the current System.Object.

   Returns: A shallow copy of the current System.Object.
  """
  pass
 def NextSample(self):
  """
  NextSample(self: PerformanceCounter) -> CounterSample

  

   Obtains a counter sample,and returns the raw,or uncalculated,value for it.

   Returns: A System.Diagnostics.CounterSample that represents the next raw value that the system obtains 

    for this counter.
  """
  pass
 def NextValue(self):
  """
  NextValue(self: PerformanceCounter) -> Single

  

   Obtains a counter sample and returns the calculated value for it.

   Returns: The next calculated value that the system obtains for this counter.
  """
  pass
 def RemoveInstance(self):
  """
  RemoveInstance(self: PerformanceCounter)

   Deletes the category instance specified by the System.Diagnostics.PerformanceCounter object 

    System.Diagnostics.PerformanceCounter.InstanceName property.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,categoryName=None,counterName=None,*__args):
  """
  __new__(cls: type)

  __new__(cls: type,categoryName: str,counterName: str,instanceName: str,machineName: str)

  __new__(cls: type,categoryName: str,counterName: str,instanceName: str)

  __new__(cls: type,categoryName: str,counterName: str,instanceName: str,readOnly: bool)

  __new__(cls: type,categoryName: str,counterName: str)

  __new__(cls: type,categoryName: str,counterName: str,readOnly: bool)
  """
  pass
 def __str__(self,*args):
  pass
 CanRaiseEvents=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the component can raise an event.



"""

 CategoryName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the name of the performance counter category for this performance counter.



Get: CategoryName(self: PerformanceCounter) -> str



Set: CategoryName(self: PerformanceCounter)=value

"""

 CounterHelp=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the description for this performance counter.



Get: CounterHelp(self: PerformanceCounter) -> str



"""

 CounterName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the name of the performance counter that is associated with this System.Diagnostics.PerformanceCounter instance.



Get: CounterName(self: PerformanceCounter) -> str



Set: CounterName(self: PerformanceCounter)=value

"""

 CounterType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the counter type of the associated performance counter.



Get: CounterType(self: PerformanceCounter) -> PerformanceCounterType



"""

 DesignMode=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the System.ComponentModel.Component is currently in design mode.



"""

 Events=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the list of event handlers that are attached to this System.ComponentModel.Component.



"""

 InstanceLifetime=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the lifetime of a process.



Get: InstanceLifetime(self: PerformanceCounter) -> PerformanceCounterInstanceLifetime



Set: InstanceLifetime(self: PerformanceCounter)=value

"""

 InstanceName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets an instance name for this performance counter.



Get: InstanceName(self: PerformanceCounter) -> str



Set: InstanceName(self: PerformanceCounter)=value

"""

 MachineName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the computer name for this performance counter



Get: MachineName(self: PerformanceCounter) -> str



Set: MachineName(self: PerformanceCounter)=value

"""

 RawValue=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the raw,or uncalculated,value of this counter.



Get: RawValue(self: PerformanceCounter) -> Int64



Set: RawValue(self: PerformanceCounter)=value

"""

 ReadOnly=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether this System.Diagnostics.PerformanceCounter instance is in read-only mode.



Get: ReadOnly(self: PerformanceCounter) -> bool



Set: ReadOnly(self: PerformanceCounter)=value

"""


 DefaultFileMappingSize=524288




class WebBrowserEncryptionLevel(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies constants that define the encryption methods used by documents displayed in the System.Windows.Forms.WebBrowser control.

 

 enum WebBrowserEncryptionLevel,values: Bit128 (6),Bit40 (3),Bit56 (4),Fortezza (5),Insecure (0),Mixed (1),Unknown (2)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Bit128=None
 Bit40=None
 Bit56=None
 Fortezza=None
 Insecure=None
 Mixed=None
 Unknown=None
 value__=None




class RenamedEventArgs(FileSystemEventArgs):
 """
 Provides data for the System.IO.FileSystemWatcher.Renamed event.

 

 RenamedEventArgs(changeType: WatcherChangeTypes,directory: str,name: str,oldName: str)
 """
 @staticmethod
 def __new__(self,changeType,directory,name,oldName):
  """ __new__(cls: type,changeType: WatcherChangeTypes,directory: str,name: str,oldName: str) """
  pass
 OldFullPath=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the previous fully qualified path of the affected file or directory.



Get: OldFullPath(self: RenamedEventArgs) -> str



"""

 OldName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the old name of the affected file or directory.



Get: OldName(self: RenamedEventArgs) -> str



"""





class ProgressBarRenderer(object):
 """ Provides methods used to render a progress bar control with visual styles. This class cannot be inherited. """
 @staticmethod
 def DrawHorizontalBar(g,bounds):
  """
  DrawHorizontalBar(g: Graphics,bounds: Rectangle)

   Draws an empty progress bar control that fills in horizontally.

  

   g: The System.Drawing.Graphics used to draw the progress bar.

   bounds: The System.Drawing.Rectangle that specifies the bounds of the progress bar.
  """
  pass
 @staticmethod
 def DrawHorizontalChunks(g,bounds):
  """
  DrawHorizontalChunks(g: Graphics,bounds: Rectangle)

   Draws a set of progress bar pieces that fill a horizontal progress bar.

  

   g: The System.Drawing.Graphics used to draw the progress bar.

   bounds: The System.Drawing.Rectangle that specifies the bounds to be filled by progress bar pieces.
  """
  pass
 @staticmethod
 def DrawVerticalBar(g,bounds):
  """
  DrawVerticalBar(g: Graphics,bounds: Rectangle)

   Draws an empty progress bar control that fills in vertically.

  

   g: The System.Drawing.Graphics used to draw the progress bar.

   bounds: The System.Drawing.Rectangle that specifies the bounds of the progress bar.
  """
  pass
 @staticmethod
 def DrawVerticalChunks(g,bounds):
  """
  DrawVerticalChunks(g: Graphics,bounds: Rectangle)

   Draws a set of progress bar pieces that fill a vertical progress bar.

  

   g: The System.Drawing.Graphics used to draw the progress bar.

   bounds: The System.Drawing.Rectangle that specifies the bounds to be filled by progress bar pieces.
  """
  pass
 ChunkSpaceThickness=0
 ChunkThickness=6
 IsSupported=True




class RectangleConverter(TypeConverter):
 """
 Converts rectangles from one data type to another. Access this class through the System.ComponentModel.TypeDescriptor.

 

 RectangleConverter()
 """
 def CanConvertFrom(self,*__args):
  """
  CanConvertFrom(self: RectangleConverter,context: ITypeDescriptorContext,sourceType: Type) -> bool

  

   Determines if this converter can convert an object in the given source type to the native type 

    of the converter.

  

  

   context: A formatter context. This object can be used to get additional information about the environment 

    this converter is being called from. This may be null,so you should always check. Also,

    properties on the context object may also return null.

  

   sourceType: The type you want to convert from.

   Returns: This method returns true if this object can perform the conversion; otherwise,false.
  """
  pass
 def CanConvertTo(self,*__args):
  """
  CanConvertTo(self: RectangleConverter,context: ITypeDescriptorContext,destinationType: Type) -> bool

  

   Gets a value indicating whether this converter can convert an object to the given destination 

    type using the context.

  

  

   context: An System.ComponentModel.ITypeDescriptorContext object that provides a format context. This can 

    be null,so you should always check. Also,properties on the context object can also return 

    null.

  

   destinationType: A System.Type object that represents the type you want to convert to.

   Returns: This method returns true if this converter can perform the conversion; otherwise,false.
  """
  pass
 def ConvertFrom(self,*__args):
  """
  ConvertFrom(self: RectangleConverter,context: ITypeDescriptorContext,culture: CultureInfo,value: object) -> object

  

   Converts the given object to a System.Drawing.Rectangle object.

  

   context: A System.ComponentModel.ITypeDescriptorContext that can be used to get additional information 

    about the environment this converter is being called from. This may be null,so you should 

    always check. Also,properties on the context object may also return null.

  

   culture: An System.Globalization.CultureInfo that contains culture specific information,such as the 

    language,calendar,and cultural conventions associated with a specific culture. It is based on 

    the RFC 1766 standard.

  

   value: The object to convert.

   Returns: The converted object.
  """
  pass
 def ConvertTo(self,*__args):
  """
  ConvertTo(self: RectangleConverter,context: ITypeDescriptorContext,culture: CultureInfo,value: object,destinationType: Type) -> object

  

   Converts the specified object to the specified type.

  

   context: A System.ComponentModel.ITypeDescriptorContext that can be used to get additional information 

    about the environment this converter is being called from. This may be null,so you should 

    always check. Also,properties on the context object may also return null.

  

   culture: An System.Globalization.CultureInfo that contains culture specific information,such as the 

    language,calendar,and cultural conventions associated with a specific culture. It is based on 

    the RFC 1766 standard.

  

   value: The object to convert.

   destinationType: The type to convert the object to.

   Returns: The converted object.
  """
  pass
 def CreateInstance(self,*__args):
  """
  CreateInstance(self: RectangleConverter,context: ITypeDescriptorContext,propertyValues: IDictionary) -> object

  

   Creates an instance of this type given a set of property values for the object. This is useful 

    for objects that are immutable but still want to provide changeable properties.

  

  

   context: A System.ComponentModel.ITypeDescriptorContext through which additional context can be provided.

   propertyValues: A dictionary of new property values. The dictionary contains a series of name-value pairs,one 

    for each property returned from a call to the 

    System.Drawing.RectangleConverter.GetProperties(System.ComponentModel.ITypeDescriptorContext,Syst

    em.Object,System.Attribute[]) method.

  

   Returns: The newly created object,or null if the object could not be created. The default implementation 

    returns null.
  """
  pass
 def GetCreateInstanceSupported(self,context=None):
  """
  GetCreateInstanceSupported(self: RectangleConverter,context: ITypeDescriptorContext) -> bool

  

   Determines if changing a value on this object should require a call to 

    System.Drawing.RectangleConverter.CreateInstance(System.ComponentModel.ITypeDescriptorContext,Sys

    tem.Collections.IDictionary) to create a new value.

  

  

   context: A type descriptor through which additional context can be provided.

   Returns: This method returns true if 

    System.Drawing.RectangleConverter.CreateInstance(System.ComponentModel.ITypeDescriptorContext,Sys

    tem.Collections.IDictionary) should be called when a change is made to one or more properties of 

    this object; otherwise,false.
  """
  pass
 def GetProperties(self,*__args):
  """
  GetProperties(self: RectangleConverter,context: ITypeDescriptorContext,value: object,attributes: Array[Attribute]) -> PropertyDescriptorCollection

  

   Retrieves the set of properties for this type. By default,a type does not return any properties.

  

   context: A System.ComponentModel.ITypeDescriptorContext through which additional context can be provided.

   value: The value of the object to get the properties for.

   attributes: An array of System.Attribute objects that describe the properties.

   Returns: The set of properties that should be exposed for this data type. If no properties should be 

    exposed,this may return null. The default implementation always returns null.
  """
  pass
 def GetPropertiesSupported(self,context=None):
  """
  GetPropertiesSupported(self: RectangleConverter,context: ITypeDescriptorContext) -> bool

  

   Determines if this object supports properties. By default,this is false.

  

   context: A System.ComponentModel.ITypeDescriptorContext through which additional context can be provided.

   Returns: This method returns true if 

    System.Drawing.RectangleConverter.GetProperties(System.ComponentModel.ITypeDescriptorContext,Syst

    em.Object,System.Attribute[]) should be called to find the properties of this object; otherwise,

    false.
  """
  pass



class _Exception:
 """ Exposes the public members of the System.Exception class to unmanaged code. """
 def Equals(self,obj):
  """
  Equals(self: _Exception,obj: object) -> bool

  

   Provides COM objects with version-independent access to the System.Object.Equals(System.Object) 

    method.

  

  

   obj: The System.Object to compare with the current System.Object.

   Returns: true if the specified System.Object is equal to the current System.Object; otherwise,false.
  """
  pass
 def GetBaseException(self):
  """
  GetBaseException(self: _Exception) -> Exception

  

   Provides COM objects with version-independent access to the System.Exception.GetBaseException 

    method.

  

   Returns: The first exception thrown in a chain of exceptions. If the System.Exception.InnerException 

    property of the current exception is a null reference (Nothing in Visual Basic),this property 

    returns the current exception.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: _Exception) -> int

  

   Provides COM objects with version-independent access to the System.Object.GetHashCode method.

   Returns: The hash code for the current instance.
  """
  pass
 def GetObjectData(self,info,context):
  """
  GetObjectData(self: _Exception,info: SerializationInfo,context: StreamingContext)

   Provides COM objects with version-independent access to the 

    System.Exception.GetObjectData(System.Runtime.Serialization.SerializationInfo,System.Runtime.Seri

    alization.StreamingContext) method

  

  

   info: The System.Runtime.Serialization.SerializationInfo object that holds the serialized object data 

    about the exception being thrown.

  

   context: The System.Runtime.Serialization.StreamingContext structure that contains contextual information 

    about the source or destination.
  """
  pass
 def GetType(self):
  """
  GetType(self: _Exception) -> Type

  

   Provides COM objects with version-independent access to the System.Exception.GetType method.

   Returns: A System.Type object that represents the exact runtime type of the current instance.
  """
  pass
 def ToString(self):
  """
  ToString(self: _Exception) -> str

  

   Provides COM objects with version-independent access to the System.Exception.ToString method.

   Returns: A string that represents the current System.Exception object.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __str__(self,*args):
  pass
 HelpLink=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Provides COM objects with version-independent access to the System.Exception.HelpLink property.



Get: HelpLink(self: _Exception) -> str



Set: HelpLink(self: _Exception)=value

"""

 InnerException=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Provides COM objects with version-independent access to the System.Exception.InnerException property.



Get: InnerException(self: _Exception) -> Exception



"""

 Message=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Provides COM objects with version-independent access to the System.Exception.Message property.



Get: Message(self: _Exception) -> str



"""

 Source=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Provides COM objects with version-independent access to the System.Exception.Source property.



Get: Source(self: _Exception) -> str



Set: Source(self: _Exception)=value

"""

 StackTrace=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Provides COM objects with version-independent access to the System.Exception.StackTrace property.



Get: StackTrace(self: _Exception) -> str



"""

 TargetSite=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Provides COM objects with version-independent access to the System.Exception.TargetSite property.



Get: TargetSite(self: _Exception) -> MethodBase



"""





class ComSourceInterfacesAttribute(Attribute,_Attribute):
 """
 Identifies a list of interfaces that are exposed as COM event sources for the attributed class.

 

 ComSourceInterfacesAttribute(sourceInterfaces: str)

 ComSourceInterfacesAttribute(sourceInterface: Type)

 ComSourceInterfacesAttribute(sourceInterface1: Type,sourceInterface2: Type)

 ComSourceInterfacesAttribute(sourceInterface1: Type,sourceInterface2: Type,sourceInterface3: Type)

 ComSourceInterfacesAttribute(sourceInterface1: Type,sourceInterface2: Type,sourceInterface3: Type,sourceInterface4: Type)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type,sourceInterfaces: str)

  __new__(cls: type,sourceInterface: Type)

  __new__(cls: type,sourceInterface1: Type,sourceInterface2: Type)

  __new__(cls: type,sourceInterface1: Type,sourceInterface2: Type,sourceInterface3: Type)

  __new__(cls: type,sourceInterface1: Type,sourceInterface2: Type,sourceInterface3: Type,sourceInterface4: Type)
  """
  pass
 Value=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the fully qualified name of the event source interface.



Get: Value(self: ComSourceInterfacesAttribute) -> str



"""





class EventDescriptor(MemberDescriptor):
 """ Provides information about an event. """
 def AddEventHandler(self,component,value):
  """
  AddEventHandler(self: EventDescriptor,component: object,value: Delegate)

   When overridden in a derived class,binds the event to the component.

  

   component: A component that provides events to the delegate.

   value: A delegate that represents the method that handles the event.
  """
  pass
 def RemoveEventHandler(self,component,value):
  """
  RemoveEventHandler(self: EventDescriptor,component: object,value: Delegate)

   When overridden in a derived class,unbinds the delegate from the component so that the delegate 

    will no longer receive events from the component.

  

  

   component: The component that the delegate is bound to.

   value: The delegate to unbind from the component.
  """
  pass
 @staticmethod
 def __new__(self,*args): #cannot find CLR constructor
  """
  __new__(cls: type,name: str,attrs: Array[Attribute])

  __new__(cls: type,descr: MemberDescriptor)

  __new__(cls: type,descr: MemberDescriptor,attrs: Array[Attribute])
  """
  pass
 AttributeArray=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets an array of attributes.



"""

 ComponentType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """When overridden in a derived class,gets the type of component this event is bound to.



Get: ComponentType(self: EventDescriptor) -> Type



"""

 EventType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """When overridden in a derived class,gets the type of delegate for the event.



Get: EventType(self: EventDescriptor) -> Type



"""

 IsMulticast=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """When overridden in a derived class,gets a value indicating whether the event delegate is a multicast delegate.



Get: IsMulticast(self: EventDescriptor) -> bool



"""

 NameHashCode=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the hash code for the name of the member,as specified in System.String.GetHashCode.



"""





class RotateFlipType(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies how much an image is rotated and the axis used to flip the image.

 

 enum RotateFlipType,values: Rotate180FlipNone (2),Rotate180FlipX (6),Rotate180FlipXY (0),Rotate180FlipY (4),Rotate270FlipNone (3),Rotate270FlipX (7),Rotate270FlipXY (1),Rotate270FlipY (5),Rotate90FlipNone (1),Rotate90FlipX (5),Rotate90FlipXY (3),Rotate90FlipY (7),RotateNoneFlipNone (0),RotateNoneFlipX (4),RotateNoneFlipXY (2),RotateNoneFlipY (6)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Rotate180FlipNone=None
 Rotate180FlipX=None
 Rotate180FlipXY=None
 Rotate180FlipY=None
 Rotate270FlipNone=None
 Rotate270FlipX=None
 Rotate270FlipXY=None
 Rotate270FlipY=None
 Rotate90FlipNone=None
 Rotate90FlipX=None
 Rotate90FlipXY=None
 Rotate90FlipY=None
 RotateNoneFlipNone=None
 RotateNoneFlipX=None
 RotateNoneFlipXY=None
 RotateNoneFlipY=None
 value__=None




"""Access loading engine properties."""
#pylint: disable=W0703,C0302,C0103,W0614,E0401,W0611,C0413,ungrouped-imports
import os.path as op
import clr

try:
    clr.AddReference('PyRevitLoader')
except Exception:
    # probably older IronPython engine not being able to
    # resolve to an already loaded assembly.
    # PyRevitLoader is executing this script so it should be referabe.
    pass

try:
    import PyRevitLoader
    ScriptExecutor = PyRevitLoader.ScriptExecutor
    EnginePrefix = ScriptExecutor.EnginePrefix
    EngineVersion = ScriptExecutor.EngineVersion
    EnginePath = op.dirname(
        clr.GetClrType(ScriptExecutor).Assembly.Location
        )
except ImportError:
    # this means that pyRevit is _not_ being loaded from a pyRevit engine
    # e.g. when importing from RevitPythonShell
    PyRevitLoader = ScriptExecutor = None
    EnginePrefix = ''
    EngineVersion = 000
    EnginePath = ''



class StatusBarPanel(Component,IComponent,IDisposable,ISupportInitialize):
 """
 Represents a panel in a System.Windows.Forms.StatusBar control. Although the System.Windows.Forms.StatusStrip control replaces and adds functionality to the System.Windows.Forms.StatusBar control of previous versions,System.Windows.Forms.StatusBar is retained for both backward compatibility and future use if you choose.

 

 StatusBarPanel()
 """
 def BeginInit(self):
  """
  BeginInit(self: StatusBarPanel)

   Begins the initialization of a System.Windows.Forms.StatusBarPanel.
  """
  pass
 def Dispose(self):
  """
  Dispose(self: StatusBarPanel,disposing: bool)

   Releases the unmanaged resources used by the System.Windows.Forms.StatusBarPanel and optionally 

    releases the managed resources.

  

  

   disposing: true to release both managed and unmanaged resources; false to release only unmanaged resources.
  """
  pass
 def EndInit(self):
  """
  EndInit(self: StatusBarPanel)

   Ends the initialization of a System.Windows.Forms.StatusBarPanel.
  """
  pass
 def GetService(self,*args):
  """
  GetService(self: Component,service: Type) -> object

  

   Returns an object that represents a service provided by the System.ComponentModel.Component or 

    by its System.ComponentModel.Container.

  

  

   service: A service provided by the System.ComponentModel.Component.

   Returns: An System.Object that represents a service provided by the System.ComponentModel.Component,or 

    null if the System.ComponentModel.Component does not provide the specified service.
  """
  pass
 def MemberwiseClone(self,*args):
  """
  MemberwiseClone(self: MarshalByRefObject,cloneIdentity: bool) -> MarshalByRefObject

  

   Creates a shallow copy of the current System.MarshalByRefObject object.

  

   cloneIdentity: false to delete the current System.MarshalByRefObject object's identity,which will cause the 

    object to be assigned a new identity when it is marshaled across a remoting boundary. A value of 

    false is usually appropriate. true to copy the current System.MarshalByRefObject object's 

    identity to its clone,which will cause remoting client calls to be routed to the remote server 

    object.

  

   Returns: A shallow copy of the current System.MarshalByRefObject object.

  MemberwiseClone(self: object) -> object

  

   Creates a shallow copy of the current System.Object.

   Returns: A shallow copy of the current System.Object.
  """
  pass
 def ToString(self):
  """
  ToString(self: StatusBarPanel) -> str

  

   Retrieves a string that contains information about the panel.

   Returns: Returns a string that contains the class name for the control and the text it contains.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __str__(self,*args):
  pass
 Alignment=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the alignment of text and icons within the status bar panel.



Get: Alignment(self: StatusBarPanel) -> HorizontalAlignment



Set: Alignment(self: StatusBarPanel)=value

"""

 AutoSize=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether the status bar panel is automatically resized.



Get: AutoSize(self: StatusBarPanel) -> StatusBarPanelAutoSize



Set: AutoSize(self: StatusBarPanel)=value

"""

 BorderStyle=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the border style of the status bar panel.



Get: BorderStyle(self: StatusBarPanel) -> StatusBarPanelBorderStyle



Set: BorderStyle(self: StatusBarPanel)=value

"""

 CanRaiseEvents=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the component can raise an event.



"""

 DesignMode=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the System.ComponentModel.Component is currently in design mode.



"""

 Events=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the list of event handlers that are attached to this System.ComponentModel.Component.



"""

 Icon=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the icon to display within the status bar panel.



Get: Icon(self: StatusBarPanel) -> Icon



Set: Icon(self: StatusBarPanel)=value

"""

 MinWidth=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the minimum allowed width of the status bar panel within the System.Windows.Forms.StatusBar control.



Get: MinWidth(self: StatusBarPanel) -> int



Set: MinWidth(self: StatusBarPanel)=value

"""

 Name=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the name to apply to the System.Windows.Forms.StatusBarPanel.



Get: Name(self: StatusBarPanel) -> str



Set: Name(self: StatusBarPanel)=value

"""

 Parent=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Windows.Forms.StatusBar control that hosts the status bar panel.



Get: Parent(self: StatusBarPanel) -> StatusBar



"""

 Style=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the style of the status bar panel.



Get: Style(self: StatusBarPanel) -> StatusBarPanelStyle



Set: Style(self: StatusBarPanel)=value

"""

 Tag=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets an object that contains data about the System.Windows.Forms.StatusBarPanel.



Get: Tag(self: StatusBarPanel) -> object



Set: Tag(self: StatusBarPanel)=value

"""

 Text=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the text of the status bar panel.



Get: Text(self: StatusBarPanel) -> str



Set: Text(self: StatusBarPanel)=value

"""

 ToolTipText=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets ToolTip text associated with the status bar panel.



Get: ToolTipText(self: StatusBarPanel) -> str



Set: ToolTipText(self: StatusBarPanel)=value

"""

 Width=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the width of the status bar panel within the System.Windows.Forms.StatusBar control.



Get: Width(self: StatusBarPanel) -> int



Set: Width(self: StatusBarPanel)=value

"""





class UnhandledExceptionEventArgs(EventArgs):
 """
 Provides data for the event that is raised when there is an exception that is not handled in any application domain.

 

 UnhandledExceptionEventArgs(exception: object,isTerminating: bool)
 """
 @staticmethod
 def __new__(self,exception,isTerminating):
  """ __new__(cls: type,exception: object,isTerminating: bool) """
  pass
 ExceptionObject=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the unhandled exception object.



Get: ExceptionObject(self: UnhandledExceptionEventArgs) -> object



"""

 IsTerminating=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Indicates whether the common language runtime is terminating.



Get: IsTerminating(self: UnhandledExceptionEventArgs) -> bool



"""





"""pyRevit managed environment variables framework.

pyRevit provides the environment variables framework to the pyRevit core
and all pyRevit tools so they can store arbitary data withing the running host
session and share small data quickly between script runs.

Some settings needs to be set for the current session and might need to affect
the behaviour of all individual scripts inside the extensions.
(e.g. If user activates the ``DEBUG`` mode, all scripts should follow and log
the debug entries.) The information is saved using ``AppDomain.GetData`` and
``SetData`` in a dictionary parameter. The dictionary is used to minimize the
addition of named parameters to the AppDomain. The dictionary then includes
all the internal parameters and their associated value. This way each script
does not need to read the usersettings data which reduces io and saves time.


pyRevit uses environment variables extensively at its core and making changes
to the core environment variables (starting with ``PYREVIT_``) through
scripts is strongly prohibited.

Examples:
    ```python
    from pyrevit.coreutils import envvars
    envvars.set_pyrevit_env_var('MY_SCRIPT_STATUS', True)
    envvars.set_pyrevit_env_var('MY_SCRIPT_CONFIG', {'someconfig': True})
    ```

    Then another script or same script when executed later within the same
    session can query the shared environment variable:

    ```python
    envvars.get_pyrevit_env_vars('MY_SCRIPT_STATUS')
    ```
    True
    ```python
    envvars.get_pyrevit_env_vars('MY_SCRIPT_CONFIG')
    ```
    {'someconfig': True}
"""

from pyrevit.framework import AppDomain
from pyrevit.labs import Common

PRODUCT_NAME = str(Common.PyRevitLabsConsts.ProductName).upper()
# root env var dictionary key.
# must be the same in this file and pyrevit/loader/runtime/envvars.cs
ENV_VAR_DICT_NAME = PRODUCT_NAME + "EnvVarsDict"

SESSIONUUID_ENVVAR = PRODUCT_NAME + "_UUID"
APPVERSION_ENVVAR = PRODUCT_NAME + "_APPVERSION"
VERSION_ENVVAR = PRODUCT_NAME + "_VERSION"
CLONENAME_ENVVAR = PRODUCT_NAME + "_CLONE"
IPYVERSION_ENVVAR = PRODUCT_NAME + "_IPYVERSION"
CPYVERSION_ENVVAR = PRODUCT_NAME + "_CPYVERSION"

LOGGING_LEVEL_ENVVAR = PRODUCT_NAME + "_LOGGINGLEVEL"
FILELOGGING_ENVVAR = PRODUCT_NAME + "_FILELOGGING"

LOADEDASSMS_ENVVAR = PRODUCT_NAME + "_LOADEDASSMS"
REFEDASSMS_ENVVAR = PRODUCT_NAME + "_REFEDASSMS"

TELEMETRYSTATE_ENVVAR = PRODUCT_NAME + "_TELEMETRYSTATE"
TELEMETRYUTCTIMESTAMPS_ENVVAR = PRODUCT_NAME + "_TELEMETRYUTCTIMESTAMPS"
TELEMETRYDIR_ENVVAR = PRODUCT_NAME + "_TELEMETRYDIR"
TELEMETRYFILE_ENVVAR = PRODUCT_NAME + "_TELEMETRYFILE"
TELEMETRYSERVER_ENVVAR = PRODUCT_NAME + "_TELEMETRYSERVER"
TELEMETRYINCLUDEHOOKS_ENVVAR = PRODUCT_NAME + "_TELEMETRYINCLUDEHOOKS"

APPTELEMETRYSTATE_ENVVAR = PRODUCT_NAME + "_APPTELEMETRYSTATE"
APPTELEMETRYHANDLER_ENVVAR = \
    PRODUCT_NAME + "_APPTELEMETRYHANDLER"
APPTELEMETRYSERVER_ENVVAR = \
    PRODUCT_NAME + "_APPTELEMETRYSERVER"
APPTELEMETRYEVENTFLAGS_ENVVAR = \
    PRODUCT_NAME + "_APPTELEMETRYEVENTFLAGS"

ROUTES_SERVER = PRODUCT_NAME + "_ROUTESSERVER"
ROUTES_ROUTES = PRODUCT_NAME + "_ROUTESROUTES"

HOOKS_ENVVAR = PRODUCT_NAME + "_HOOKS"
HOOKSHANDLER_ENVVAR = PRODUCT_NAME + "_HOOKSHANDLER"

AUTOUPDATING_ENVVAR = PRODUCT_NAME + "_AUTOUPDATE"
OUTPUT_STYLESHEET_ENVVAR = PRODUCT_NAME + "_STYLESHEET"
RIBBONUPDATOR_ENVVAR = PRODUCT_NAME + "_RIBBONUPDATOR"
TABCOLORIZER_ENVVAR = PRODUCT_NAME + "_TABCOLORIZER"


def get_pyrevit_env_vars():
    """Get the root dictionary, holding all environment variables."""
    return AppDomain.CurrentDomain.GetData(ENV_VAR_DICT_NAME)


def get_pyrevit_env_var(param_name):
    """Get value of a parameter shared between all scripts.

    Args:
        param_name (str): name of environment variable

    Returns:
        (object): any object stored as the environment variable value
    """
    # This function returns None if it can not find the parameter.
    # Thus value of None should not be used for params

    data_dict = get_pyrevit_env_vars()
    return data_dict.get(param_name) if data_dict else None


def set_pyrevit_env_var(param_name, param_value):
    """Set value of a parameter shared between all scripts.

    Args:
        param_name (str): name of environment variable
        param_value (object): any python object
    """
    # Get function returns None if it can not find the parameter.
    # Thus value of None should not be used for params
    data_dict = get_pyrevit_env_vars() or {}
    data_dict[param_name] = param_value
    AppDomain.CurrentDomain.SetData(ENV_VAR_DICT_NAME, data_dict)



"""Hooks management."""
import os.path as op
import re
from collections import namedtuple

from pyrevit import HOST_APP
from pyrevit import framework
from pyrevit import coreutils
from pyrevit.coreutils.logger import get_logger
from pyrevit.coreutils import envvars
from pyrevit.runtime.types import EventHooks
import pyrevit.extensions as exts

from pyrevit.loader import sessioninfo


SUPPORTED_LANGUAGES = [
    exts.PYTHON_SCRIPT_FILE_FORMAT,
    exts.CSHARP_SCRIPT_FILE_FORMAT,
    exts.VB_SCRIPT_FILE_FORMAT,
    ]

#pylint: disable=W0703,C0302,C0103
mlogger = get_logger(__name__)


ExtensionEventHook = namedtuple('ExtensionEventHook', [
    'id',
    'name',
    'target',
    'script',
    'syspaths',
    'extension_name',
    ])


def get_hooks_handler():
    """Get the hook handler environment variable.

    Returns:
        (EventHooks): hook handler
    """
    return envvars.get_pyrevit_env_var(envvars.HOOKSHANDLER_ENVVAR)


def set_hooks_handler(handler):
    """Set the hook handler environment variable.

    Args:
        handler (EventHooks): hook handler
    """
    envvars.set_pyrevit_env_var(envvars.HOOKSHANDLER_ENVVAR, handler)


def is_valid_hook_script(hook_script):
    """Check if the given hook script is valid.

    Args:
        hook_script (str): hook script path

    Returns:
        (bool): True if the script is valid, False otherwise
    """
    return op.splitext(op.basename(hook_script))[1] in SUPPORTED_LANGUAGES


def _get_hook_parts(extension, hook_script):
    # finds the two parts of the hook script name
    # e.g command-before-exec[ID_INPLACE_COMPONENT].py
    # ('command-before-exec', 'ID_INPLACE_COMPONENT')
    parts = re.findall(
        r'([a-z -]+)\[?([A-Z _]+)?\]?\..+',
        op.basename(hook_script)
        )
    if parts:
        return parts[0]
    else:
        return '', ''


def _create_hook_id(extension, hook_script):
    hook_script_id = op.basename(hook_script)
    pieces = [extension.unique_name, hook_script_id]
    return coreutils.cleanup_string(
        exts.UNIQUE_ID_SEPARATOR.join(pieces),
        skip=[exts.UNIQUE_ID_SEPARATOR]
        ).lower()


def get_extension_hooks(extension):
    """Get the hooks of the given extension.

    Args:
        extension (pyrevit.extensions.components.Extension): pyRevit extension

    Returns:
        (list[ExtensionEventHook]): list of hooks
    """
    event_hooks = []
    for hook_script in extension.get_hooks():
        if is_valid_hook_script(hook_script):
            name, target = _get_hook_parts(extension, hook_script)
            if name:
                event_hooks.append(
                    ExtensionEventHook(
                        id=_create_hook_id(extension, hook_script),
                        name=name,
                        target=target,
                        script=hook_script,
                        syspaths=extension.module_paths,
                        extension_name=extension.name,
                    )
                )
    return event_hooks


def get_event_hooks():
    """Get all the event hooks."""
    hooks_handler = get_hooks_handler()
    return hooks_handler.GetAllEventHooks()


def register_hooks(extension):
    """Register the hooks for the given extension.

    Args:
        extension (pyrevit.extensions.components.Extension): pyRevit extension
    """
    hooks_handler = get_hooks_handler()
    for ext_hook in get_extension_hooks(extension):
        try:
            hooks_handler.RegisterHook(
                uniqueId=ext_hook.id,
                eventName=ext_hook.name,
                eventTarget=ext_hook.target,
                scriptPath=ext_hook.script,
                searchPaths=framework.Array[str](ext_hook.syspaths),
                extensionName=ext_hook.extension_name,
            )
        except Exception as hookEx:
            mlogger.error("Failed registering hook script %s | %s",
                          ext_hook.script, hookEx)


def unregister_hooks(extension):
    """Unregister all hooks for the given extension.

    Args:
        extension (pyrevit.extensions.components.Extension): pyRevit extension
    """
    hooks_handler = get_hooks_handler()
    for ext_hook in get_extension_hooks(extension):
        hooks_handler.UnRegisterHook(uniqueId=ext_hook.id)


def unregister_all_hooks():
    """Unregister all hooks."""
    hooks_handler = get_hooks_handler()
    hooks_handler.UnRegisterAllHooks(uiApp=HOST_APP.uiapp)


def activate():
    """Activate all event hooks."""
    hooks_handler = get_hooks_handler()
    hooks_handler.ActivateEventHooks(uiApp=HOST_APP.uiapp)


def deactivate():
    """Deactivate all event hooks."""
    hooks_handler = get_hooks_handler()
    hooks_handler.DeactivateEventHooks(uiApp=HOST_APP.uiapp)


def setup_hooks(session_id=None):
    """Setup the hooks for the given session.
    
    If no session is specified, use the current one.

    Args:
        session_id (str, optional): Session. Defaults to None.
    """
    # make sure session id is availabe
    if not session_id:
        session_id = sessioninfo.get_session_uuid()

    hooks_handler = get_hooks_handler()
    if hooks_handler:
        # deactivate old
        hooks_handler.DeactivateEventHooks(uiApp=HOST_APP.uiapp)
    # setup new
    hooks_handler = EventHooks(session_id)
    set_hooks_handler(hooks_handler)
    unregister_all_hooks()



"""Revit Server API commands and dictionary keys."""

# http request header
REQ_HEADER_USERNAME = 'User-Name'
REQ_HEADER_MACHINE = 'User-Machine-Name'
REQ_HEADER_GUID = 'Operation-GUID'

# folder structure
DIVIDER = '|'

# commands
# ------------------------------------------------------------------------------
REQ_CMD_SERVERPROP = "/serverproperties"
# {
#     "AccessLevelTypes": list,
#     "MachineName": str,
#     "MaximumFolderPathLength": int,
#     "MaximumModelNameLength": int,
#     "ServerRoles": list,
#     "Servers": list
# }
SERVER_ACCESSLEVEL_KEY = "AccessLevelTypes"
SERVER_MACHINENAME_KEY = "MachineName"
SERVER_MAXPATHLENGTH_KEY = "MaximumFolderPathLength"
SERVER_MAXNAMELENGTH_KEY = "MaximumModelNameLength"
SERVER_ROLES_KEY = "ServerRoles"
SERVER_SERVERS_KEY = "Servers"

# ------------------------------------------------------------------------------
REQ_CMD_CONTENTS = "/contents"
# {
#     "Path": str,
#     "DriveFreeSpace": int,
#     "DriveSpace": int,
#     "Files": list,
#     "Folders": list,
#     "LockContext": str,
#     "LockState": int,
#     "ModelLocksInProgress": list,
#     "Models": list
# }
NODE_PATH_KEY = "Path"
NODE_DRIVE_FREESPACE_KEY = "DriveFreeSpace"
NODE_DRIVE_TOTALSPACE_KEY = "DriveSpace"
NODE_FILES_KEY = "Files"
NODE_FOLDERS_KEY = "Folders"
NODE_LOCK_CTX_KEY = "LockContext"
NODE_LOCK_STATE_KEY = "LockState"
NODE_LOCKS_INPROGRESS_KEY = "ModelLocksInProgress"
NODE_MODELS_KEY = "Models"

# "Files" key
# {
#     "IsText": bool,
#     "Name": str,
#     "Size": int
# }
NODE_FILES_ISTXT_KEY = "IsText"
NODE_FILES_NAME_KEY = "Name"
NODE_FILES_SIZE_KEY = "Size"

# "Folders" key
# {
#     "HasContents": bool,
#     "LockContext": str,
#     "LockState": int,
#     "ModelLocksInProgress": list,
#     "Name": str,
#     "Size": int
# }
NODE_FOLDERS_HASCONTENTS_KEY = "HasContents"
NODE_FOLDERS_LOCKCONTEXT_KEY = "LockContext"
NODE_FOLDERS_LOCKSTATE_KEY = "LockState"
NODE_FOLDERS_LOCKINPROGRESS_KEY = "ModelLocksInProgress"
NODE_FOLDERS_NAME_KEY = "Name"
NODE_FOLDERS_SIZE_KEY = "Size"

# "Models" key
# {
#     "LockContext": str,
#     "LockState": int,
#     "ModelLocksInProgress": list,
#     "ModelSize": int,
#     "Name": str,
#     "ProductVersion": int,
#     "SupportSize": int
# }
NODE_MODELS_LOCKCONTEXT_KEY = "LockContext"
NODE_MODELS_LOCKSTATE_KEY = "LockState"
NODE_MODELS_LOCKINPROGRESS_KEY = "ModelLocksInProgress"
NODE_MODELS_SIZE_KEY = "ModelSize"
NODE_MODELS_NAME_KEY = "Name"
NODE_MODELS_PRODUCTVERSION_KEY = "ProductVersion"
NODE_MODELS_SUPPORTSIZE_KEY = "SupportSize"


# ------------------------------------------------------------------------------
REQ_CMD_DIRINFO = "/directoryinfo"
# {
#     "Path": str,
#     "DateCreated": str, # "/Date(1483465123167)/"
#     "DateModified": str, # "/Date(1501619985043)/"
#     "Exists": bool,
#     "FolderCount": int,
#     "IsFolder": bool,
#     "LastModifiedBy": str,
#     "LockContext": str,
#     "LockState": int,
#     "ModelCount": int,
#     "ModelLocksInProgress": list,
#     "ModelSize": int,
#     "Size": int
# }
NODE_DIRINFO_PATH_KEY = "Path"
NODE_DIRINFO_DATECREATED_KEY = "DateCreated"
NODE_DIRINFO_DATEMODIFIED_KEY = "DateModified"
NODE_DIRINFO_EXISTS_KEY = "Exists"
NODE_DIRINFO_FOLDERCOUNT_KEY = "FolderCount"
NODE_DIRINFO_ISFOLDER_KEY = "IsFolder"
NODE_DIRINFO_LASTMODIFIEDBY_KEY = "LastModifiedBy"
NODE_DIRINFO_LOCKCTX_KEY = "LockContext"
NODE_DIRINFO_LOCKSTATE_KEY = "LockState"
NODE_DIRINFO_MODELCOUNT_KEY = "ModelCount"
NODE_DIRINFO_LOCKSINPROGRESS_KEY = "ModelLocksInProgress"
NODE_DIRINFO_MODELSIZE_KEY = "ModelSize"
NODE_DIRINFO_SIZE_KEY = "Size"


# ------------------------------------------------------------------------------
REQ_CMD_MODELINFO = "/modelinfo"
# {
#     "Path": str,
#     "DateCreated": str, #"/Date(1483465179783)/"
#     "DateModified": str, # "/Date(1503702490000)/"
#     "LastModifiedBy": str,
#     "ModelGUID": str, # "b04725c0-9369-4482-aecf-5ad900d4c1bb"
#     "ModelSize": int,
#     "SupportSize": int
# }
NODE_MODELINFO_PATH_KEY = "Path"
NODE_MODELINFO_DATECREATED_KEY = "DateCreated"
NODE_MODELINFO_DATEMODIFIED_KEY = "DateModified"
NODE_MODELINFO_LASTMODIFIEDBY_KEY = "LastModifiedBy"
NODE_MODELINFO_MODELGUID_KEY = "ModelGUID"
NODE_MODELINFO_MODELSIZE_KEY = "ModelSize"
NODE_MODELINFO_SUPPORTSIZE_KEY = "SupportSize"


# ------------------------------------------------------------------------------
REQ_CMD_PROJINFO = "/projectinfo"
# [
#     {
#         "A:title": "Text",
#         "<>": dict
#     }
# ]
PARAM_CATNAME_KEY = "A:title"

# Parameter dict keys
# {
#     "#text": "BLDG",
#     "@displayName": "Building Name",
#     "@id": "940f16f5-879a-4bf6-b722-b6a0043ffa9b",
#     "@type": "system",
#     "@typeOfParameter": "Text"
# }
PARAM_VALUE_KEY = "#text"
PARAM_NAME_KEY = "@displayName"
PARAM_ID_KEY = "@id"
PARAM_TYPE_KEY = "@type"
PARAM_DTYPE_KEY = "@typeOfParameter"

# ------------------------------------------------------------------------------
REQ_CMD_MHISTORY = "/history"
# {
#     "Path": str,
#     "Items": list
# }
MHISTORY_PATH_KEY = "Path"
MHISTORY_ITEMS_KEY = "Items"

# {
#     "Comment": str,
#     "Date": str, # "/Date(1483465201000)/"
#     "ModelSize": int,
#     "OverwrittenByHistoryNumber": int,
#     "SupportSize": int,
#     "User": str,
#     "VersionNumber": int
# }
MHISTORYITEM_COMMENT_KEY = "Comment"
MHISTORY_DATE_KEY = "Date"
MHISTORY_MODELSIZE_KEY = "ModelSize"
MHISTORY_OVERWRITE_KEY = "OverwrittenByHistoryNumber"
MHISTORY_SUPPORTSIZE_KEY = "SupportSize"
MHISTORY_USER_KEY = "User"
MHISTORY_VERSION_KEY = "VersionNumber"

# ------------------------------------------------------------------------------
# shared data types
# "ModelLocksInProgress" key
# {
#     "Age": str, # "PT29M32.1723042S"
#     "ModelLockOptions": int,
#     "ModelLockType": int,
#     "ModelPath": str,
#     "TimeStamp": str, # "/Date(1504130384000)/"
#     "UserName": str
# }
NODE_LIP_AGE_KEY = "Age"
NODE_LIP_LOCKOPTIONS_KEY = "ModelLockOptions"
NODE_LIP_LOCKTYPE_KEY = "ModelLockType"
NODE_LIP_MODELPATH_KEY = "ModelPath"
NODE_LIP_TIMESTAMP_KEY = "TimeStamp"
NODE_LIP_USERNAME_KEY = "UserName"


# ------------------------------------------------------------------------------
REQ_CMD_LOCK = "/lock"
# ------------------------------------------------------------------------------
REQ_CMD_CANCELLOCK = "/inProgressLock"
# ------------------------------------------------------------------------------
REQ_CMD_UNLOCK = "/lock?objectMustExist=true"
# ------------------------------------------------------------------------------
REQ_CMD_CHILDNLOCKS = "/descendent/locks"
# {
#     "Path": str,
#     "Items": list
#     "DescendentHasLockContext": bool
# }
CHILDLOCKS_PATH_KEY = "Path"
CHILDLOCKS_ITEMS_KEY = "Items"
CHILDLOCKS_LOCKCTX = "DescendentHasLockContext"

# {
#     "Path": str,
#     "FailedItems": list
# }
CHILDLOCKS_DELPATH_KEY = "Path"
CHILDLOCKS_DELFAILEDITEMS_KEY = "FailedItems"

# ------------------------------------------------------------------------------
REQ_CMD_MKDIR = ""
# ------------------------------------------------------------------------------
REQ_CMD_DELETE = "?newObjectName"
# ------------------------------------------------------------------------------
REQ_CMD_RENAME = "?newObjectName={new_name}"
# ------------------------------------------------------------------------------
REQ_CMD_COPY = "?destinationObjectPath={dest_path}"\
               "&pasteAction=Copy"\
               "&replaceExisting={replace_exist}"
# ------------------------------------------------------------------------------
REQ_CMD_MOVE = "?destinationObjectPath={dest_path}"\
               "&pasteAction=Move"\
               "&replaceExisting={replace_exist}"



class IWebProxy:
 """ Provides the base interface for implementation of proxy access for the System.Net.WebRequest class. """
 def GetProxy(self,destination):
  """
  GetProxy(self: IWebProxy,destination: Uri) -> Uri

  

   Returns the URI of a proxy.

  

   destination: A System.Uri that specifies the requested Internet resource.

   Returns: A System.Uri instance that contains the URI of the proxy used to contact destination.
  """
  pass
 def IsBypassed(self,host):
  """
  IsBypassed(self: IWebProxy,host: Uri) -> bool

  

   Indicates that the proxy should not be used for the specified host.

  

   host: The System.Uri of the host to check for proxy use.

   Returns: true if the proxy server should not be used for host; otherwise,false.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 Credentials=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """The credentials to submit to the proxy server for authentication.



Get: Credentials(self: IWebProxy) -> ICredentials



Set: Credentials(self: IWebProxy)=value

"""





class ListSortDescriptionCollection(object,IList,ICollection,IEnumerable):
 """
 Represents a collection of System.ComponentModel.ListSortDescription objects.

 

 ListSortDescriptionCollection()

 ListSortDescriptionCollection(sorts: Array[ListSortDescription])
 """
 def Contains(self,value):
  """
  Contains(self: ListSortDescriptionCollection,value: object) -> bool

  

   Determines if the System.ComponentModel.ListSortDescriptionCollection contains a specific value.

  

   value: The System.Object to locate in the collection.

   Returns: true if the System.Object is found in the collection; otherwise,false.
  """
  pass
 def CopyTo(self,array,index):
  """
  CopyTo(self: ListSortDescriptionCollection,array: Array,index: int)

   Copies the contents of the collection to the specified array,starting at the specified 

    destination array index.

  

  

   array: The destination array for the items copied from the collection.

   index: The index of the destination array at which copying begins.
  """
  pass
 def IndexOf(self,value):
  """
  IndexOf(self: ListSortDescriptionCollection,value: object) -> int

  

   Returns the index of the specified item in the collection.

  

   value: The System.Object to locate in the collection.

   Returns: The index of value if found in the list; otherwise,-1.
  """
  pass
 def __contains__(self,*args):
  """
  __contains__(self: IList,value: object) -> bool

  

   Determines whether the System.Collections.IList contains a specific value.

  

   value: The object to locate in the System.Collections.IList.

   Returns: true if the System.Object is found in the System.Collections.IList; otherwise,false.
  """
  pass
 def __getitem__(self,*args):
  """ x.__getitem__(y) <==> x[y] """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 def __len__(self,*args):
  """ x.__len__() <==> len(x) """
  pass
 @staticmethod
 def __new__(self,sorts=None):
  """
  __new__(cls: type)

  __new__(cls: type,sorts: Array[ListSortDescription])
  """
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 def __setitem__(self,*args):
  """ x.__setitem__(i,y) <==> x[i]= """
  pass
 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of items in the collection.



Get: Count(self: ListSortDescriptionCollection) -> int



"""





class IProgress:
 # no doc
 def Report(self,value):
  """ Report(self: IProgress[T],value: T) """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class WindowsFormsSection(ConfigurationSection):
 """
 Defines a new System.Configuration.ConfigurationSection for parsing application settings. This class cannot be inherited.

 

 WindowsFormsSection()
 """
 ElementProperty=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Configuration.ConfigurationElementProperty object that represents the System.Configuration.ConfigurationElement object itself.



"""

 EvaluationContext=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Configuration.ContextInformation object for the System.Configuration.ConfigurationElement object.



"""

 HasContext=property(lambda self: object(),lambda self,v: None,lambda self: None)

 JitDebugging=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether just-in-time (JIT) debugging is used.



Get: JitDebugging(self: WindowsFormsSection) -> bool



Set: JitDebugging(self: WindowsFormsSection)=value

"""

 Properties=property(lambda self: object(),lambda self,v: None,lambda self: None)





class StyleSelector(object):
 """
 Provides a way to apply styles based on custom logic.

 

 StyleSelector()
 """
 def SelectStyle(self,item,container):
  """
  SelectStyle(self: StyleSelector,item: object,container: DependencyObject) -> Style

  

   When overridden in a derived class,returns a System.Windows.Style based on custom logic.

  

   item: The content.

   container: The element to which the style will be applied.

   Returns: Returns an application-specific style to apply; otherwise,null.
  """
  pass



class Double(object):
 """ Represents a double-precision floating-point number. """
 def as_integer_ratio(self,*args):
  """ as_integer_ratio(self: float) -> tuple """
  pass
 def conjugate(self,*args):
  """ conjugate(x: float) -> float """
  pass
 def fromhex(self,*args):
  """ fromhex(cls: type,self: str) -> object """
  pass
 def hex(self,*args):
  """ hex(self: float) -> str """
  pass
 def is_integer(self,*args):
  """ is_integer(self: float) -> bool """
  pass
 def __abs__(self,*args):
  """ x.__abs__() <==> abs(x) """
  pass
 def __add__(self,*args):
  """ x.__add__(y) <==> x+y """
  pass
 def __coerce__(self,*args):
  """ __coerce__(x: float,o: object) -> tuple """
  pass
 def __divmod__(self,*args):
  """ __divmod__(x: float,y: float) -> object """
  pass
 def __div__(self,*args):
  """ x.__div__(y) <==> x/y """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __float__(self,*args):
  """ __float__(self: float) -> float """
  pass
 def __floordiv__(self,*args):
  """ x.__floordiv__(y) <==> x//y """
  pass
 def __getformat__(self,*args):
  """ __getformat__(typestr: str) -> str """
  pass
 def __getnewargs__(self,*args):
  """ __getnewargs__(self: float) -> object """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __int__(self,*args):
  """ __int__(d: float) -> object """
  pass
 def __le__(self,*args):
  pass
 def __long__(self,*args):
  """ __long__(self: float) -> long """
  pass
 def __lt__(self,*args):
  pass
 def __mod__(self,*args):
  """ x.__mod__(y) <==> x%y """
  pass
 def __mul__(self,*args):
  """ x.__mul__(y) <==> x*y """
  pass
 def __neg__(self,*args):
  """ x.__neg__() <==> -x """
  pass
 @staticmethod
 def __new__(self,*args): #cannot find CLR constructor
  """
  __new__(cls: type,s: IList[Byte]) -> object

  __new__(cls: type,x: object) -> object

  __new__(cls: type) -> object
  """
  pass
 def __ne__(self,*args):
  pass
 def __nonzero__(self,*args):
  """ __nonzero__(x: float) -> bool """
  pass
 def __pos__(self,*args):
  """ __pos__(x: float) -> float """
  pass
 def __pow__(self,*args):
  """ x.__pow__(y[,z]) <==> pow(x,y[,z]) """
  pass
 def __radd__(self,*args):
  """ __radd__(x: float,y: float) -> float """
  pass
 def __rdivmod__(self,*args):
  """ __rdivmod__(x: float,y: float) -> object """
  pass
 def __rdiv__(self,*args):
  """ __rdiv__(x: float,y: float) -> float """
  pass
 def __rfloordiv__(self,*args):
  """ __rfloordiv__(x: float,y: float) -> float """
  pass
 def __rmod__(self,*args):
  """ __rmod__(x: float,y: float) -> float """
  pass
 def __rmul__(self,*args):
  """ __rmul__(x: float,y: float) -> float """
  pass
 def __rpow__(self,*args):
  """ __rpow__(x: float,y: float) -> float """
  pass
 def __rsub__(self,*args):
  """ __rsub__(x: float,y: float) -> float """
  pass
 def __rtruediv__(self,*args):
  """ __rtruediv__(x: float,y: float) -> float """
  pass
 def __setformat__(self,*args):
  """ __setformat__(typestr: str,fmt: str) """
  pass
 def __sub__(self,*args):
  """ x.__sub__(y) <==> x-y """
  pass
 def __truediv__(self,*args):
  """ x.__truediv__(y) <==> x/y """
  pass
 def __trunc__(self,*args):
  """ __trunc__(x: float) -> object """
  pass
 imag=None
 real=None




class IRecyclingItemContainerGenerator(IItemContainerGenerator):
 """ Extends the System.Windows.Controls.Primitives.IItemContainerGenerator interface to reuse the UI content it generates. Classes that are responsible for generating user interface (UI) content on behalf of a host implement this interface. """
 def Recycle(self,position,count):
  """
  Recycle(self: IRecyclingItemContainerGenerator,position: GeneratorPosition,count: int)

   Disassociates item containers from their data items and saves the containers so they can be 

    reused later for other data items.

  

  

   position: The zero-based index of the first element to reuse. position must refer to a previously 

    generated (realized) item.

  

   count: The number of elements to reuse,starting at position.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class AutomationProxyAttribute(Attribute,_Attribute):
 """
 Specifies whether the type should be marshaled using the Automation marshaler or a custom proxy and stub.

 

 AutomationProxyAttribute(val: bool)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,val):
  """ __new__(cls: type,val: bool) """
  pass
 Value=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating the type of marshaler to use.



Get: Value(self: AutomationProxyAttribute) -> bool



"""





class Byte(object,IComparable,IFormattable,IConvertible,IComparable[Byte],IEquatable[Byte]):
 """ Represents an 8-bit unsigned integer. """
 def bit_length(self,*args):
  """ bit_length(value: Byte) -> int """
  pass
 def CompareTo(self,value):
  """
  CompareTo(self: Byte,value: Byte) -> int

  

   Compares this instance to a specified 8-bit unsigned integer and returns an indication of their 

    relative values.

  

  

   value: An 8-bit unsigned integer to compare.

   Returns: A signed integer that indicates the relative order of this instance and value.Return Value 

    Description Less than zero This instance is less than value. Zero This instance is equal to 

    value. Greater than zero This instance is greater than value.

  

  CompareTo(self: Byte,value: object) -> int

  

   Compares this instance to a specified object and returns an indication of their relative values.

  

   value: An object to compare,or null.

   Returns: A signed integer that indicates the relative order of this instance and value.Return Value 

    Description Less than zero This instance is less than value. Zero This instance is equal to 

    value. Greater than zero This instance is greater than value.-or- value is null.
  """
  pass
 def conjugate(self,*args):
  """ conjugate(x: Byte) -> Byte """
  pass
 def Equals(self,obj):
  """
  Equals(self: Byte,obj: Byte) -> bool

  

   Returns a value indicating whether this instance and a specified System.Byte object represent 

    the same value.

  

  

   obj: An object to compare to this instance.

   Returns: true if obj is equal to this instance; otherwise,false.

  Equals(self: Byte,obj: object) -> bool

  

   Returns a value indicating whether this instance is equal to a specified object.

  

   obj: An object to compare with this instance,or null.

   Returns: true if obj is an instance of System.Byte and equals the value of this instance; otherwise,

    false.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: Byte) -> int

  

   Returns the hash code for this instance.

   Returns: A hash code for the current System.Byte.
  """
  pass
 def GetTypeCode(self):
  """
  GetTypeCode(self: Byte) -> TypeCode

  

   Returns the System.TypeCode for value type System.Byte.

   Returns: The enumerated constant,System.TypeCode.Byte.
  """
  pass
 @staticmethod
 def Parse(s,*__args):
  """
  Parse(s: str,provider: IFormatProvider) -> Byte

  

   Converts the string representation of a number in a specified culture-specific format to its 

    System.Byte equivalent.

  

  

   s: A string that contains a number to convert. The string is interpreted using the 

    System.Globalization.NumberStyles.Integer style.

  

   provider: An object that supplies culture-specific parsing information about s. If provider is null,the 

    thread current culture is used.

  

   Returns: A byte value that is equivalent to the number contained in s.

  Parse(s: str,style: NumberStyles,provider: IFormatProvider) -> Byte

  

   Converts the string representation of a number in a specified style and culture-specific format 

    to its System.Byte equivalent.

  

  

   s: A string that contains a number to convert. The string is interpreted using the style specified 

    by style.

  

   style: A bitwise combination of enumeration values that indicates the style elements that can be 

    present in s. A typical value to specify is System.Globalization.NumberStyles.Integer.

  

   provider: An object that supplies culture-specific information about the format of s. If provider is null,

    the thread current culture is used.

  

   Returns: A byte value that is equivalent to the number contained in s.

  Parse(s: str) -> Byte

  

   Converts the string representation of a number to its System.Byte equivalent.

  

   s: A string that contains a number to convert. The string is interpreted using the 

    System.Globalization.NumberStyles.Integer style.

  

   Returns: A byte value that is equivalent to the number contained in s.

  Parse(s: str,style: NumberStyles) -> Byte

  

   Converts the string representation of a number in a specified style to its System.Byte 

    equivalent.

  

  

   s: A string that contains a number to convert. The string is interpreted using the style specified 

    by style.

  

   style: A bitwise combination of enumeration values that indicates the style elements that can be 

    present in s. A typical value to specify is System.Globalization.NumberStyles.Integer.

  

   Returns: A byte value that is equivalent to the number contained in s.
  """
  pass
 def ToString(self,*__args):
  """
  ToString(self: Byte,provider: IFormatProvider) -> str

  

   Converts the numeric value of the current System.Byte object to its equivalent string 

    representation using the specified culture-specific formatting information.

  

  

   provider: An object that supplies culture-specific formatting information.

   Returns: The string representation of the value of this object in the format specified by the provider 

    parameter.

  

  ToString(self: Byte,format: str,provider: IFormatProvider) -> str

  

   Converts the value of the current System.Byte object to its equivalent string representation 

    using the specified format and culture-specific formatting information.

  

  

   format: A standard or custom numeric format string.

   provider: An object that supplies culture-specific formatting information.

   Returns: The string representation of the current System.Byte object,formatted as specified by the 

    format and provider parameters.

  

  ToString(self: Byte) -> str

  

   Converts the value of the current System.Byte object to its equivalent string representation.

   Returns: The string representation of the value of this object,which consists of a sequence of digits 

    that range from 0 to 9 with no leading zeroes.

  

  ToString(self: Byte,format: str) -> str

  

   Converts the value of the current System.Byte object to its equivalent string representation 

    using the specified format.

  

  

   format: A numeric format string.

   Returns: The string representation of the current System.Byte object,formatted as specified by the 

    format parameter.
  """
  pass
 @staticmethod
 def TryParse(s,*__args):
  """
  TryParse(s: str,style: NumberStyles,provider: IFormatProvider) -> (bool,Byte)

  

   Converts the string representation of a number in a specified style and culture-specific format 

    to its System.Byte equivalent. A return value indicates whether the conversion succeeded or 

    failed.

  

  

   s: A string containing a number to convert. The string is interpreted using the style specified by 

    style.

  

   style: A bitwise combination of enumeration values that indicates the style elements that can be 

    present in s. A typical value to specify is System.Globalization.NumberStyles.Integer.

  

   provider: An object that supplies culture-specific formatting information about s. If provider is null,

    the thread current culture is used.

  

   Returns: true if s was converted successfully; otherwise,false.

  TryParse(s: str) -> (bool,Byte)

  

   Tries to convert the string representation of a number to its System.Byte equivalent,and 

    returns a value that indicates whether the conversion succeeded.

  

  

   s: A string that contains a number to convert. The string is interpreted using the 

    System.Globalization.NumberStyles.Integer style.

  

   Returns: true if s was converted successfully; otherwise,false.
  """
  pass
 def __abs__(self,*args):
  """ x.__abs__() <==> abs(x) """
  pass
 def __add__(self,*args):
  """ x.__add__(y) <==> x+yx.__add__(y) <==> x+y """
  pass
 def __and__(self,*args):
  """
  __and__(x: Byte,y: SByte) -> Int16

  __and__(x: Byte,y: Byte) -> Byte
  """
  pass
 def __cmp__(self,*args):
  """ x.__cmp__(y) <==> cmp(x,y)x.__cmp__(y) <==> cmp(x,y) """
  pass
 def __div__(self,*args):
  """ x.__div__(y) <==> x/yx.__div__(y) <==> x/y """
  pass
 def __float__(self,*args):
  """ __float__(x: Byte) -> float """
  pass
 def __floordiv__(self,*args):
  """ x.__floordiv__(y) <==> x//yx.__floordiv__(y) <==> x//y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __hash__(self,*args):
  """ x.__hash__() <==> hash(x) """
  pass
 def __hex__(self,*args):
  """ __hex__(value: Byte) -> str """
  pass
 def __index__(self,*args):
  """ __index__(x: Byte) -> int """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __int__(self,*args):
  """ __int__(x: Byte) -> int """
  pass
 def __invert__(self,*args):
  """ __invert__(x: Byte) -> object """
  pass
 def __lshift__(self,*args):
  """ x.__rshift__(y) <==> x<<yx.__rshift__(y) <==> x<<y """
  pass
 def __mod__(self,*args):
  """ x.__mod__(y) <==> x%yx.__mod__(y) <==> x%y """
  pass
 def __mul__(self,*args):
  """ x.__mul__(y) <==> x*yx.__mul__(y) <==> x*y """
  pass
 def __neg__(self,*args):
  """ x.__neg__() <==> -x """
  pass
 @staticmethod
 def __new__(self,*args): #cannot find CLR constructor
  """
  __new__(cls: type,value: object) -> object

  __new__(cls: type) -> object
  """
  pass
 def __nonzero__(self,*args):
  """ __nonzero__(x: Byte) -> bool """
  pass
 def __or__(self,*args):
  """
  __or__(x: Byte,y: SByte) -> Int16

  __or__(x: Byte,y: Byte) -> Byte
  """
  pass
 def __pos__(self,*args):
  """ __pos__(x: Byte) -> Byte """
  pass
 def __pow__(self,*args):
  """ x.__pow__(y[,z]) <==> pow(x,y[,z])x.__pow__(y[,z]) <==> pow(x,y[,z]) """
  pass
 def __radd__(self,*args):
  """
  __radd__(x: SByte,y: Byte) -> object

  __radd__(x: Byte,y: Byte) -> object
  """
  pass
 def __rand__(self,*args):
  """
  __rand__(x: SByte,y: Byte) -> Int16

  __rand__(x: Byte,y: Byte) -> Byte
  """
  pass
 def __rdiv__(self,*args):
  """
  __rdiv__(x: SByte,y: Byte) -> object

  __rdiv__(x: Byte,y: Byte) -> object
  """
  pass
 def __reduce_ex__(self,*args):
  pass
 def __repr__(self,*args):
  """ __repr__(x: Byte) -> str """
  pass
 def __rfloordiv__(self,*args):
  """
  __rfloordiv__(x: SByte,y: Byte) -> object

  __rfloordiv__(x: Byte,y: Byte) -> Byte
  """
  pass
 def __rmod__(self,*args):
  """
  __rmod__(x: SByte,y: Byte) -> Int16

  __rmod__(x: Byte,y: Byte) -> Byte
  """
  pass
 def __rmul__(self,*args):
  """
  __rmul__(x: SByte,y: Byte) -> object

  __rmul__(x: Byte,y: Byte) -> object
  """
  pass
 def __ror__(self,*args):
  """
  __ror__(x: SByte,y: Byte) -> Int16

  __ror__(x: Byte,y: Byte) -> Byte
  """
  pass
 def __rpow__(self,*args):
  """
  __rpow__(x: SByte,y: Byte) -> object

  __rpow__(x: Byte,y: Byte) -> object
  """
  pass
 def __rshift__(self,*args):
  """ x.__rshift__(y) <==> x>>yx.__rshift__(y) <==> x>>y """
  pass
 def __rsub__(self,*args):
  """
  __rsub__(x: SByte,y: Byte) -> object

  __rsub__(x: Byte,y: Byte) -> object
  """
  pass
 def __rtruediv__(self,*args):
  """
  __rtruediv__(x: SByte,y: Byte) -> float

  __rtruediv__(x: Byte,y: Byte) -> float
  """
  pass
 def __rxor__(self,*args):
  """
  __rxor__(x: SByte,y: Byte) -> Int16

  __rxor__(x: Byte,y: Byte) -> Byte
  """
  pass
 def __str__(self,*args):
  pass
 def __sub__(self,*args):
  """ x.__sub__(y) <==> x-yx.__sub__(y) <==> x-y """
  pass
 def __truediv__(self,*args):
  """ x.__truediv__(y) <==> x/yx.__truediv__(y) <==> x/y """
  pass
 def __trunc__(self,*args):
  """ __trunc__(x: Byte) -> Byte """
  pass
 def __xor__(self,*args):
  """
  __xor__(x: Byte,y: SByte) -> Int16

  __xor__(x: Byte,y: Byte) -> Byte
  """
  pass
 denominator=None
 imag=None
 MaxValue=None
 MinValue=None
 numerator=None
 real=None




class CookieCollection(object,ICollection,IEnumerable):
 """
 Provides a collection container for instances of the System.Net.Cookie class.

 

 CookieCollection()
 """
 def Add(self,*__args):
  """
  Add(self: CookieCollection,cookies: CookieCollection)

   Adds the contents of a System.Net.CookieCollection to the current instance.

  

   cookies: The System.Net.CookieCollection to be added.

  Add(self: CookieCollection,cookie: Cookie)

   Adds a System.Net.Cookie to a System.Net.CookieCollection.

  

   cookie: The System.Net.Cookie to be added to a System.Net.CookieCollection.
  """
  pass
 def CopyTo(self,array,index):
  """
  CopyTo(self: CookieCollection,array: Array[Cookie],index: int)

   Copies the elements of this System.Net.CookieCollection to a System.Net.Cookie array starting at 

    the specified index of the target array.

  

  

   array: The target System.Net.Cookie array to which the System.Net.CookieCollection will be copied.

   index: The zero-based index in the target System.Array where copying begins.

  CopyTo(self: CookieCollection,array: Array,index: int)

   Copies the elements of a System.Net.CookieCollection to an instance of the System.Array class,

    starting at a particular index.

  

  

   array: The target System.Array to which the System.Net.CookieCollection will be copied.

   index: The zero-based index in the target System.Array where copying begins.
  """
  pass
 def GetEnumerator(self):
  """
  GetEnumerator(self: CookieCollection) -> IEnumerator

  

   Gets an enumerator that can iterate through a System.Net.CookieCollection.

   Returns: An instance of an implementation of an System.Collections.IEnumerator interface that can iterate 

    through a System.Net.CookieCollection.
  """
  pass
 def __add__(self,*args):
  """ x.__add__(y) <==> x+yx.__add__(y) <==> x+y """
  pass
 def __getitem__(self,*args):
  """ x.__getitem__(y) <==> x[y]x.__getitem__(y) <==> x[y] """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 def __len__(self,*args):
  """ x.__len__() <==> len(x) """
  pass
 def __reduce_ex__(self,*args):
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of cookies contained in a System.Net.CookieCollection.



Get: Count(self: CookieCollection) -> int



"""

 IsReadOnly=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether a System.Net.CookieCollection is read-only.



Get: IsReadOnly(self: CookieCollection) -> bool



"""

 IsSynchronized=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether access to a System.Net.CookieCollection is thread safe.



Get: IsSynchronized(self: CookieCollection) -> bool



"""

 SyncRoot=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets an object that you can use to synchronize access to the System.Net.CookieCollection.



Get: SyncRoot(self: CookieCollection) -> object



"""





"""Utility fuctions to support smart modules."""
import inspect
import types


def copy_func(func, func_name, doc_string=None, arg_list=None):
    """Copy a function object to create a new function.

    This is used inside smart modules that auto-generate functions based on
    context.

    Args:
        func (object): python source function object
        func_name (str): new function name
        doc_string (str): new function docstring
        arg_list (list): list of default values for function arguments

    Returns:
        (object): new python function objects
    """
    new_func = types.FunctionType(func.func_code, func.func_globals,
                                  func_name, tuple(arg_list), func.func_closure)
    new_func.__doc__ = doc_string
    return new_func


def mark(prop_name):
    """Decorator function to add a marker property to the given type."""
    def setter_decorator(type_obj):
        setattr(type_obj, prop_name, True)
        return type_obj
    return setter_decorator


def collect_marked(module_obj, prop_name):
    """Collect module objects that are marked with given property."""
    marked_objs = []
    for member in inspect.getmembers(module_obj):
        _, type_obj = member
        if (inspect.isclass(type_obj) or inspect.isfunction(type_obj)) \
                and getattr(type_obj, prop_name, False):
            marked_objs.append(type_obj)
    return marked_objs


def has_argument(function_obj, arg_name):
    """Check if given function object has argument matching arg_name."""
    return arg_name in inspect.getargspec(function_obj)[0] #pylint: disable=deprecated-method


def has_any_arguments(function_obj, arg_name_list):
    """Check if given function object has any of given arguments."""
    args = inspect.getargspec(function_obj)[0] #pylint: disable=deprecated-method
    if arg_name_list:
        return any(x in args for x in arg_name_list)
    return False


def filter_kwargs(function_obj, kwargs):
    """Filter given arguments dict for function_obj arguments."""
    filtered_kwargs = {}
    for arg_name in inspect.getargspec(function_obj)[0]: #pylint: disable=deprecated-method
        filtered_kwargs[arg_name] = kwargs.get(arg_name, None)
    return filtered_kwargs



class PARAMFLAG(Enum,IComparable,IFormattable,IConvertible):
 """
 Use System.Runtime.InteropServices.ComTypes.PARAMFLAG instead.

 

 enum (flags) PARAMFLAG,values: PARAMFLAG_FHASCUSTDATA (64),PARAMFLAG_FHASDEFAULT (32),PARAMFLAG_FIN (1),PARAMFLAG_FLCID (4),PARAMFLAG_FOPT (16),PARAMFLAG_FOUT (2),PARAMFLAG_FRETVAL (8),PARAMFLAG_NONE (0)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 PARAMFLAG_FHASCUSTDATA=None
 PARAMFLAG_FHASDEFAULT=None
 PARAMFLAG_FIN=None
 PARAMFLAG_FLCID=None
 PARAMFLAG_FOPT=None
 PARAMFLAG_FOUT=None
 PARAMFLAG_FRETVAL=None
 PARAMFLAG_NONE=None
 value__=None




class DateBoldEventArgs(EventArgs):
 """ Provides data for events that are internal to the System.Windows.Forms.MonthCalendar control. """
 DaysToBold=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets dates that are bold.



Get: DaysToBold(self: DateBoldEventArgs) -> Array[int]



Set: DaysToBold(self: DateBoldEventArgs)=value

"""

 Size=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of dates that are bold.



Get: Size(self: DateBoldEventArgs) -> int



"""

 StartDate=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the first date that is bold.



Get: StartDate(self: DateBoldEventArgs) -> DateTime



"""





class HierarchicalVirtualizationConstraints(object):
 """ HierarchicalVirtualizationConstraints(cacheLength: VirtualizationCacheLength,cacheLengthUnit: VirtualizationCacheLengthUnit,viewport: Rect) """
 def Equals(self,*__args):
  """
  Equals(self: HierarchicalVirtualizationConstraints,comparisonConstraints: HierarchicalVirtualizationConstraints) -> bool

  Equals(self: HierarchicalVirtualizationConstraints,oCompare: object) -> bool
  """
  pass
 def GetHashCode(self):
  """ GetHashCode(self: HierarchicalVirtualizationConstraints) -> int """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 @staticmethod
 def __new__(self,cacheLength,cacheLengthUnit,viewport):
  """
  __new__(cls: type,cacheLength: VirtualizationCacheLength,cacheLengthUnit: VirtualizationCacheLengthUnit,viewport: Rect)

  __new__[HierarchicalVirtualizationConstraints]() -> HierarchicalVirtualizationConstraints
  """
  pass
 def __ne__(self,*args):
  pass
 CacheLength=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: CacheLength(self: HierarchicalVirtualizationConstraints) -> VirtualizationCacheLength



"""

 CacheLengthUnit=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: CacheLengthUnit(self: HierarchicalVirtualizationConstraints) -> VirtualizationCacheLengthUnit



"""

 Viewport=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: Viewport(self: HierarchicalVirtualizationConstraints) -> Rect



"""





class ICustomFactory:
 """ Enables users to write activation code for managed objects that extend System.MarshalByRefObject. """
 def CreateInstance(self,serverType):
  """
  CreateInstance(self: ICustomFactory,serverType: Type) -> MarshalByRefObject

  

   Creates a new instance of the specified type.

  

   serverType: The type to activate.

   Returns: A System.MarshalByRefObject associated with the specified type.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class DataGridViewEditingControlShowingEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.DataGridView.EditingControlShowing event.

 

 DataGridViewEditingControlShowingEventArgs(control: Control,cellStyle: DataGridViewCellStyle)
 """
 @staticmethod
 def __new__(self,control,cellStyle):
  """ __new__(cls: type,control: Control,cellStyle: DataGridViewCellStyle) """
  pass
 CellStyle=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the cell style of the edited cell.



Get: CellStyle(self: DataGridViewEditingControlShowingEventArgs) -> DataGridViewCellStyle



Set: CellStyle(self: DataGridViewEditingControlShowingEventArgs)=value

"""

 Control=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """The control shown to the user for editing the selected cell's value.



Get: Control(self: DataGridViewEditingControlShowingEventArgs) -> Control



"""





# -*- coding: utf-8 -*-
r"""Python wrapper for Autodesk Revit Server.

Example:
    ```python
    name = '<server name>'
    version = '2017'    # server version in XXXX format
    rserver = RevitServer(name, version)
    # listing all files, folders, and models in a server
    for parent, folders, files, models in rserver.walk():
        print(parent)
        for fd in folders:
            print('\t@d {}'.format(fd.path))
        for f in files:
            print('\t@f {}'.format(f.path))
        for m in models:
            print('\t@m {}'.format(m.path))
    ```
"""

import os.path as op
import uuid
import getpass
import socket
import urllib

# third party modules
import requests

# rpws components
import rpws
import rpws.api as api
import rpws.models as models


# Default portion of the server url. The only odd one is 2012 which does not
# include server version in url. Each Revit server version can only host
# Revit models of the same version. Also it's practical for a company to want
# a consistent server name to host models of different Revit version.
# Thus the server version was added to the url after 2012 version.
# Otherwise the url for two servers called X for 2016 and 2017 Revit models
# would have been the same url.
sroots = {"2012": "/RevitServerAdminRESTService/AdminRESTService.svc",
          "2013": "/RevitServerAdminRESTService2013/AdminRESTService.svc",
          "2014": "/RevitServerAdminRESTService2014/AdminRESTService.svc",
          "2015": "/RevitServerAdminRESTService2015/AdminRESTService.svc",
          "2016": "/RevitServerAdminRESTService2016/AdminRESTService.svc",
          "2017": "/RevitServerAdminRESTService2017/AdminRESTService.svc",
          "2018": "/RevitServerAdminRESTService2018/AdminRESTService.svc",
          "2019": "/RevitServerAdminRESTService2019/AdminRESTService.svc",
          "2020": "/RevitServerAdminRESTService2020/AdminRESTService.svc",
          "2021": "/RevitServerAdminRESTService2021/AdminRESTService.svc",
          "2022": "/RevitServerAdminRESTService2022/AdminRESTService.svc",
          "2023": "/RevitServerAdminRESTService2023/AdminRESTService.svc",
          "2024": "/RevitServerAdminRESTService2024/AdminRESTService.svc",
          "2025": "/RevitServerAdminRESTService2025/AdminRESTService.svc",
          "2026": "/RevitServerAdminRESTService2026/AdminRESTService.svc",
          }


class RevitServer(object):
    """Handles all communication with Revit Server as initialized.

    Args:
        name (str): Server name.
        version (str): Server version.
        username (str, optional): Username to be passed to in http calls.
                                  This is set to current user if not provided.
        machine (str, optional): Machine name to be passed to in http calls
                                 Set to current machine if not provided.

    Attributes:
        name (str): Server name.
        version (str): Server version.
        _base_uri (str, private): Base URI of the initialized server
        _huser (str, private): Username of the initialized server.
        _hmachine (str, private): Machine name of the initialized server

    Example:
        >>> rserver = RevitServer('server01', '2017')
        >>> print(rserver)
        <rpws.RevitServer name:server01 version:2017>
    """

    def __init__(self, name, version, username=None, machine=None):
        """Class constructor."""
        # make sure version is of type str
        if type(version) == int:
            version = str(version)

        # verify server version is supported
        if version not in sroots:
            raise rpws.ServerVersionNotSupported(
                'Supported versions are: {}'
                .format([x for x in sroots.keys()]))

        self.name = name
        self.version = version
        self._base_uri = "http://" + self.name + sroots[self.version]

        if username:
            self._huser = username
        else:
            self._huser = getpass.getuser()

        if machine:
            self._hmachine = machine
        else:
            self._hmachine = socket.gethostname()

    def __repr__(self):
        """Repr for RevitServer object.

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> print(rserver)
            <rpws.RevitServer name:server01 version:2017>
        """
        return '<rpws.{} name:{} version:{}>'\
               .format(self.__class__.__name__, self.name, self.version)

    @property
    def _header_dict(self):
        """Private property that creates and returns the http header dict.

        Returns:
            dict: Header dict to be passed to Revit Server
        """
        # using uuid module to generate a unique session id that is required
        # by the Revit Server for logging purposes
        return {api.REQ_HEADER_USERNAME: self._huser,
                api.REQ_HEADER_MACHINE: self._hmachine,
                api.REQ_HEADER_GUID: str(uuid.uuid4())}

    def _httpmethod(self, http_method, command, node_uri=None, rootcmd=False):
        """The single method that handles all http requests.

        Args:
            http_method (func): method from requests module (e.g. requests.get)
            command (str): Revit Server API command from rpws.api
            node_uri (str, optional): Path to an entry on the server.
                                      Root if not provided.
            rootcmd (bool, optional): True if command should be executed
                                      at root level. Default is False.

        Returns: Depending on the input command returns
            int: http response status code
            dict: data dict returned by server as the command result

        Raises:
            rpws.ServerNotImplemented: When server does not exist
            rpws.ServerFileNotFound: When file, folder, or model does not exist
            rpws.ServerForbiddenError: On accessing forbidden entries
            rpws.ServerInternalError: When server has internal error
            rpws.ServerMethodNotAllowedError: When method is not allowed
            rpws.ServerURITooLongError: When url is too long for server
            rpws.ServerTimeoutError: On connection timeout
            rpws.ServerConnectionError: On connection error
            rpws.ServerBadRequestError: On bad request url
            rpws.ServerServiceUnavailableError: When service is not available
            rpws.UnhandledException: Any other http status codes
        """
        # create the http request url
        if rootcmd:
            req_url = self._base_uri + command
        else:
            req_url = self._base_uri + '/' + urllib.quote(self._api_path(node_uri).encode('utf-8')) + command

        # send to server
        try:
            r = http_method(req_url, headers=self._header_dict)
        except requests.ConnectTimeout:
            raise rpws.ServerTimeoutError(req_url)
        except requests.ConnectionError:
            raise rpws.ServerConnectionError(self._base_uri)

        # process status codes and results
        if r.status_code == 200 and r.text:
            r.encoding = 'utf-8-sig'
            return r.json()
        elif 200 <= r.status_code < 300:
            return True
        elif r.status_code == 400:
            raise rpws.ServerBadRequestError(node_uri)
        elif r.status_code == 403:
            raise rpws.ServerForbiddenError(node_uri)
        elif r.status_code == 404:
            raise rpws.ServerFileNotFound(node_uri)
        elif r.status_code == 405:
            raise rpws.ServerMethodNotAllowedError(node_uri)
        elif r.status_code == 414:
            raise rpws.ServerURITooLongError(node_uri)
        elif r.status_code == 500:
            raise rpws.ServerInternalError(node_uri)
        elif r.status_code == 501:
            raise rpws.ServerNotImplemented('name:{} version:{}'
                                            .format(self.name, self.version))
        elif r.status_code == 503:
            raise rpws.ServerServiceUnavailableError(node_uri)
        else:
            raise rpws.UnhandledException('requests status code:{}'
                                          .format(r.status_code))

    def _get(self, command, node_uri=None, rootcmd=False):
        """Send a GET request to Revit Server.

        Args:
            command (str): Revit Server API command from rpws.api
            node_uri (str, optional): Path to an entry on the server.
                                      Root if not provided.
            rootcmd (bool, optional): True if command should be executed
                                      at root level. Default is False.

        Returns:
            See _httpmethod() method for results
        """
        return self._httpmethod(requests.get, command, node_uri, rootcmd)

    def _post(self, command, node_uri=None, rootcmd=False):
        """Send a POST request to Revit Server.

        Args:
            command (str): Revit Server API command from rpws.api
            node_uri (str, optional): Path to an entry on the server.
                                      Root if not provided.
            rootcmd (bool, optional): True if command should be executed
                                      at root level. Default is False.

        Returns:
            See _httpmethod() method for results
        """
        return self._httpmethod(requests.post, command, node_uri, rootcmd)

    def _put(self, command, node_uri=None, rootcmd=False):
        """Send a PUT request to Revit Server.

        Args:
            command (str): Revit Server API command from rpws.api
            node_uri (str, optional): Path to an entry on the server.
                                      Root if not provided.
            rootcmd (bool, optional): True if command should be executed
                                      at root level. Default is False.

        Returns:
            See _httpmethod() method for results
        """
        return self._httpmethod(requests.put, command, node_uri, rootcmd)

    def _delete(self, command, node_uri=None, rootcmd=False):
        """Send a DELETE request to Revit Server.

        Args:
            command (str): Revit Server API command from rpws.api
            node_uri (str, optional): Path to an entry on the server.
                                      Root if not provided.
            rootcmd (bool, optional): True if command should be executed
                                      at root level. Default is False.

        Returns:
            See _httpmethod() method for results
        """
        return self._httpmethod(requests.delete, command, node_uri, rootcmd)

    @staticmethod
    def _api_path(nodepath=None):
        """Format path for Revit server requests uri.

        Generate a Revit Server directory structure path from provided
        file, folder, or model path. Revit server uses '|' to separate
        directory entries so as not to conflict with '/' in http urls. So:
        "/Training/MyOffice/2017/Model.rvt" will be changed to:
        "|Training|MyOffice|2017|Model.rvt"

        Args:
            nodepath (str, optional): Path to an entry on the server.
                                      Root if not provided.

        Returns:
            str: Path formatted for Revit Server urls
        """
        if nodepath:
            return nodepath.replace(op.sep, api.DIVIDER)
        else:
            return api.DIVIDER

    @staticmethod
    def _root_path(nodepath=None):
        """Make sure that the path starts with / as root of the server.

        Args:
            nodepath (str, optional): Path to an entry on the server.
                                      Root if not provided.

        Returns:
            str: Reformatted path
        """
        if nodepath:
            npath = op.normpath(nodepath)
            return npath if npath.startswith(op.sep) \
                else op.join(op.sep, npath)
        else:
            return op.sep

    @staticmethod
    def _getserverdriveinfo(contents_dict):
        """Return drive space info acquired from server.

        Data keys: "DriveFreeSpace" and "DriveSpace"

        Args:
            contents_dict (dict): Data dict returned by server

        Returns:
            rpws.models.ServerDriveInfo
        """
        # make the server drive info obj
        return models.ServerDriveInfo(
            drive_space=contents_dict[api.NODE_DRIVE_TOTALSPACE_KEY],
            drive_freespace=contents_dict[api.NODE_DRIVE_FREESPACE_KEY])

    @staticmethod
    def _getlocks(ip_lock_list):
        """Extract locks from list of locks in progress returned by server.

        Args:
            ip_lock_list (list): List of dict returned by server representing
                             locks in progress.

        Returns:
            rpws.models.IPLockInfo
        """
        locks_list = []

        # check to make sure ip_lock_list is not None
        if ip_lock_list:
            # for each lock in progress dict
            for ip_lock in ip_lock_list:

                # extract and create timestamp obj
                ts = models.DateEntry.\
                    fromrsdatestring(ip_lock[api.NODE_LIP_TIMESTAMP_KEY])

                # extract and create timespan obj
                tspan = models.TimeSpanEntry.\
                    fromrstimespanstring(ip_lock[api.NODE_LIP_AGE_KEY])

                # extract and create lock options and type
                lop = models.LockOptions(ip_lock[api.NODE_LIP_LOCKOPTIONS_KEY])
                lt = models.LockType(ip_lock[api.NODE_LIP_LOCKTYPE_KEY])

                # extract and create in-progress lock info obj
                locks_list.append(
                    models.IPLockInfo(
                        age=tspan,
                        lock_options=lop,
                        lock_type=lt,
                        model_path=ip_lock[api.NODE_LIP_MODELPATH_KEY],
                        timestamp=ts,
                        username=ip_lock[api.NODE_LIP_USERNAME_KEY]))

        return locks_list

    def _getfiles(self, nodepath, contents_dict):
        """Extract and return the list of all files under provided path.

        Args:
            nodepath (str): Path to an entry on the server.
            contents_dict (dict): data dict from server

        Returns:
            list of rpws.models.FileInfo
        """
        return [models.FileInfo(
            path=op.join(nodepath if nodepath else self.path,
                         x[api.NODE_FILES_NAME_KEY]),
            name=x[api.NODE_FILES_NAME_KEY],
            size=x[api.NODE_FILES_SIZE_KEY],
            is_text=x[api.NODE_FILES_ISTXT_KEY])
            for x in contents_dict.get(api.NODE_FILES_KEY, [])]

    def _getfolders(self, nodepath, contents_dict):
        """Extract and return the list of subfolders.

        Args:
            nodepath (str): Path to an entry on the server.
            contents_dict (dict): data dict from server

        Returns:
            list of rpws.models.FolderInfo
        """
        folder_infos = []

        for fdict in contents_dict.get(api.NODE_FOLDERS_KEY, []):
            # get in-progress lock objs
            ip_locks = self._getlocks(
                fdict[api.NODE_FOLDERS_LOCKINPROGRESS_KEY])
            # make lock state obj
            lock_state = \
                models.LockState(fdict[api.NODE_FOLDERS_LOCKSTATE_KEY])
            # create folder info obj
            finfo = models.FolderInfo(
                path=op.join(nodepath if nodepath else self.path,
                             fdict[api.NODE_FOLDERS_NAME_KEY]),
                name=fdict[api.NODE_FOLDERS_NAME_KEY],
                size=fdict[api.NODE_FOLDERS_SIZE_KEY],
                has_contents=fdict[api.NODE_FOLDERS_HASCONTENTS_KEY],
                lock_context=fdict[api.NODE_FOLDERS_LOCKCONTEXT_KEY],
                lock_state=lock_state,
                locks_inprogress=ip_locks)

            folder_infos.append(finfo)

        return folder_infos

    def _getmodels(self, nodepath, contents_dict):
        """Extract and return the list of models in source.

        Args:
            nodepath (str): Path to an entry on the server.
            contents_dict (dict): data dict from server

        Returns:
            list of rpws.models.ModelInfo
        """
        model_infos = []
        for mdict in contents_dict.get(api.NODE_MODELS_KEY, []):
            # get in-progress lock objs
            ip_locks = \
                self._getlocks(mdict[api.NODE_MODELS_LOCKINPROGRESS_KEY])
            # make lock state obj
            lock_state = models.LockState(mdict[api.NODE_MODELS_LOCKSTATE_KEY])
            # create model info obj
            minfo = models.ModelInfo(
                path=op.join(nodepath if nodepath else self.path,
                             mdict[api.NODE_MODELS_NAME_KEY]),
                name=mdict[api.NODE_MODELS_NAME_KEY],
                size=mdict[api.NODE_MODELS_SIZE_KEY],
                support_size=mdict[api.NODE_MODELS_SUPPORTSIZE_KEY],
                product_version=mdict[api.NODE_MODELS_PRODUCTVERSION_KEY],
                lock_context=mdict[api.NODE_MODELS_LOCKCONTEXT_KEY],
                lock_state=lock_state,
                locks_inprogress=ip_locks)

            model_infos.append(minfo)

        return model_infos

    @property
    def path(self):
        """Root path of server.

        Returns:
            str: Root path
        """
        return op.sep

    def getinfo(self):
        """Return server properties.

        API command:
            /serverproperties

        Returns:
            rpws.models.ServerInfo

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> sinfo = rserver.getinfo()
        """
        # get properties dict from server on root
        props_dict = self._get(api.REQ_CMD_SERVERPROP, rootcmd=True)

        # server roles
        sroles = [models.ServerRole(x)
                  for x in props_dict[api.SERVER_ROLES_KEY]]

        # make the server info obj
        return models.ServerInfo(
            name=self.name,
            version=self.version,
            machine_name=props_dict[api.SERVER_MACHINENAME_KEY],
            roles=sroles,
            access_level_types=props_dict[api.SERVER_ACCESSLEVEL_KEY],
            max_path_length=props_dict[api.SERVER_MAXPATHLENGTH_KEY],
            max_name_length=props_dict[api.SERVER_MAXNAMELENGTH_KEY],
            servers=props_dict[api.SERVER_SERVERS_KEY])

    def getdriveinfo(self):
        """Return server drive information.

        API command:
            /contents

        API keys:
            DriveFreeSpace and DriveSpace

        Returns:
            rpws.models.ServerDriveInfo

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> dinfo = rserver.getdriveinfo()
            >>> print(dinfo.drive_space)
        """
        return self._getserverdriveinfo(self._get(api.REQ_CMD_CONTENTS))

    def scandir(self, nodepath=None):
        """Return files, folders, and models from root or provided path.

        API command:
            /contents

        Args:
            nodepath (str, optional): Path to an entry on the server.
                                      Root if not provided.

        Returns:
            rpws.models.EntryContents

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for entry in rserver.scandir('/example/path'):
            ...     print(entry.path)
        """
        # get the entry contents (root if nodepath is none
        contents_dict = self._get(api.REQ_CMD_CONTENTS, nodepath)

        # get drive info
        sdriveinfo = self._getserverdriveinfo(contents_dict)

        # get the files, folders, and models
        node_files = self._getfiles(nodepath, contents_dict)
        node_folders = self._getfolders(nodepath, contents_dict)
        node_models = self._getmodels(nodepath, contents_dict)

        lock_ctx = contents_dict[api.NODE_LOCK_CTX_KEY]
        lock_state = models.LockState(contents_dict[api.NODE_LOCK_STATE_KEY])
        ip_locks = self._getlocks(contents_dict[api.NODE_LOCKS_INPROGRESS_KEY])

        return models.EntryContents(
            path=nodepath,
            drive_space=sdriveinfo.drive_space,
            drive_freespace=sdriveinfo.drive_freespace,
            files=node_files,
            folders=node_folders,
            lock_context=lock_ctx,
            lock_state=lock_state,
            locks_inprogress=ip_locks,
            models=node_models)

    def listfiles(self, nodepath=None):
        """Return files from root or provided path.

        API command:
            /contents

        API key:
            Files

        Args:
            nodepath (str, optional): Path to an entry on the server.
                                      Root if not provided.

        Returns:
            list of rpws.models.FileInfo

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for file in rserver.listfiles('/example/path'):
            ...     print(file.path)
        """
        return self._getfiles(nodepath,
                              self._get(api.REQ_CMD_CONTENTS, nodepath))

    def listfolders(self, nodepath=None):
        """Return folders from root or provided path.

        API command:
            /contents

        API key:
            Folders

        Args:
            nodepath (str, optional): Path to an entry on the server.
                                      Root if not provided.

        Returns:
            list of rpws.models.FolderInfo

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for folder in rserver.listfolders('/example/path'):
            ...     print(folder.path)
        """
        return self._getfolders(nodepath,
                                self._get(api.REQ_CMD_CONTENTS, nodepath))

    def listmodels(self, nodepath=None):
        """Return models from root or provided path.

        API command:
            /contents

        API key:
            Models

        Args:
            nodepath (str, optional): Path to an entry on the server.
                                      Root if not provided.

        Returns:
            list of rpws.models.ModelInfo

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for model in rserver.listmodels('/example/path'):
            ...     print(model.path)
        """
        return self._getmodels(nodepath,
                               self._get(api.REQ_CMD_CONTENTS, nodepath))

    def getfolderinfo(self, nodepath):
        """Return directory info for provided path.

        API command:
            /directoryinfo

        Args:
            nodepath (str): Path to an entry on the server.

        Returns:
            rpws.models.EntryDirInfo

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for folder in rserver.listfolders('/example/path'):
            ...     finfo = rserver.getfolderinfo(folder.path)
            ...     print(finfo.date_created)
        """
        if nodepath:
            # directory info from the entry
            ddict = self._get(api.REQ_CMD_DIRINFO, nodepath)

            # in-progress locks
            ip_locks = self._getlocks(
                ddict[api.NODE_DIRINFO_LOCKSINPROGRESS_KEY])

            # make lock state
            lock_state = \
                models.LockState(ddict[api.NODE_DIRINFO_LOCKSTATE_KEY])

            # make time stamps
            date_created = models.DateEntry.\
                fromrsdatestring(ddict[api.NODE_DIRINFO_DATECREATED_KEY])
            date_modified = models.DateEntry.\
                fromrsdatestring(ddict[api.NODE_DIRINFO_DATEMODIFIED_KEY])

            # make the directory info obj
            return models.EntryDirInfo(
                path=nodepath,
                name=op.basename(nodepath),
                size=ddict[api.NODE_DIRINFO_SIZE_KEY],
                date_created=date_created,
                date_modified=date_modified,
                exists=ddict[api.NODE_DIRINFO_EXISTS_KEY],
                folder_count=ddict[api.NODE_DIRINFO_FOLDERCOUNT_KEY],
                is_folder=ddict[api.NODE_DIRINFO_ISFOLDER_KEY],
                last_modified_by=ddict[api.NODE_DIRINFO_LASTMODIFIEDBY_KEY],
                lock_context=ddict[api.NODE_DIRINFO_LOCKCTX_KEY],
                lock_state=lock_state,
                model_count=ddict[api.NODE_DIRINFO_MODELCOUNT_KEY],
                model_size=ddict[api.NODE_DIRINFO_MODELSIZE_KEY],
                locks_inprogress=ip_locks)

    def getmodelinfo(self, nodepath):
        """Return model info from provided model path.

        API command:
            /modelinfo

        Args:
            nodepath (str): Path to a model on the server.

        Returns:
            rpws.models.ModelInfoEx

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for model in rserver.listmodels('/example/path'):
            ...     minfo = rserver.getmodelinfo(model.path)
            ...     print(minfo.size)
        """
        if nodepath:
            # model info from the entry
            mdict = self._get(api.REQ_CMD_MODELINFO, nodepath)

            # make time stamps
            dc = models.DateEntry.\
                fromrsdatestring(mdict[api.NODE_MODELINFO_DATECREATED_KEY])
            dm = models.DateEntry.\
                fromrsdatestring(mdict[api.NODE_MODELINFO_DATEMODIFIED_KEY])

            # make the model info obj
            return models.ModelInfoEx(
                path=nodepath,
                name=op.basename(nodepath),
                size=mdict[api.NODE_MODELINFO_MODELSIZE_KEY],
                guid=mdict[api.NODE_MODELINFO_MODELGUID_KEY],
                date_created=dc,
                date_modified=dm,
                last_modified_by=mdict[api.NODE_MODELINFO_LASTMODIFIEDBY_KEY],
                support_size=mdict[api.NODE_MODELINFO_SUPPORTSIZE_KEY])

    def getmodelhistory(self, nodepath):
        """Return model info from provided model path.

        API command:
            /history

        Args:
            nodepath (str): Path to a model on the server.

        Returns:
            rpws.models.MHistoryInfo

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for model in rserver.listmodels('/example/path'):
            ...     mhist = rserver.getmodelhistory(model.path)
            ...     for hist in mhist.items:
            ...        print(hist.comment, hist.user)
        """
        # get history data from server
        mhist_dict = self._get(api.REQ_CMD_MHISTORY, nodepath)

        hist_items = []

        for hitem_dict in mhist_dict[api.MHISTORY_ITEMS_KEY]:
            # make time stamp
            mhist_date = models.DateEntry.\
                fromrsdatestring(hitem_dict[api.MHISTORY_DATE_KEY])

            # make model history item info
            mhist = models.MHistoryItemInfo(
                id=hitem_dict[api.MHISTORY_VERSION_KEY],
                comment=hitem_dict[api.MHISTORYITEM_COMMENT_KEY],
                date=mhist_date,
                model_size=hitem_dict[api.MHISTORY_MODELSIZE_KEY],
                overwrittenby=hitem_dict[api.MHISTORY_OVERWRITE_KEY],
                support_size=hitem_dict[api.MHISTORY_SUPPORTSIZE_KEY],
                user=hitem_dict[api.MHISTORY_USER_KEY])

            hist_items.append(mhist)

        # make model history info obj
        return models.MHistoryInfo(path=mhist_dict[api.MHISTORY_PATH_KEY],
                                   items=hist_items)

    def getprojectinfo(self, nodepath):
        """Return project info from provided model path.

        API command:
            /projectinfo

        Args:
            nodepath (str): Path to an entry on the server.

        Returns:
            rpws.models.ProjectInfo

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for model in rserver.listmodels('/example/path'):
            ...     pinfo = rserver.getprojectinfo(model.path)
            ...     for pparam in pinfo.paramters:
            ...        print(pparam.name, pparam.value)
        """
        param_list = []

        if nodepath:
            # get all parameter categories from server for this model
            param_cats = self._get(api.REQ_CMD_PROJINFO, nodepath)

            for cat in param_cats:
                # for each category create parameter obj for its parameters
                catname = cat[api.PARAM_CATNAME_KEY]
                cat.pop(api.PARAM_CATNAME_KEY)

                for pname, param in cat.items():
                    # get the type string for the property
                    # and setup param type obj
                    ptypestr = param.get(api.PARAM_TYPE_KEY, '').lower()
                    if ptypestr:
                        ptype = models.ParamType(ptypestr)
                    else:
                        ptype = models.ParamType.Unknown

                    # get the datatype string for the property
                    # and setup param data type obj
                    pdatatypestr = param.get(api.PARAM_DTYPE_KEY, '').lower()
                    if pdatatypestr:
                        pdatatype = models.ParamDataType(pdatatypestr)
                    else:
                        pdatatype = models.ParamDataType.Unknown

                    # make the project parameter obj
                    pparam = models.ProjParameter(
                        name=param.get(api.PARAM_NAME_KEY, ''),
                        value=param.get(api.PARAM_VALUE_KEY, ''),
                        id=param.get(api.PARAM_ID_KEY, ''),
                        category=catname,
                        type=ptype,
                        datatype=pdatatype)

                    param_list.append(pparam)

        return models.ProjectInfo(param_list)

    def lock(self, nodepath):
        """Lock model.

        Args:
            nodepath (str): Path to a model on the server.

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for model in rserver.listmodels('/example/path'):
            ...     rserver.lock(model.path)
        """
        return self._put(api.REQ_CMD_LOCK, nodepath)

    def cancellock(self, nodepath):
        """Cancel any in-progress locks on model.

        Args:
            nodepath (str): Path to a model on the server.

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for model in rserver.listmodels('/example/path'):
            ...     rserver.cancellock(model.path)
        """
        return self._delete(api.REQ_CMD_CANCELLOCK, nodepath)

    def unlock(self, nodepath):
        """Unlock model.

        Args:
            nodepath (str): Path to a model on the server.

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for model in rserver.listmodels('/example/path'):
            ...     rserver.unlock(model.path)
        """
        return self._delete(api.REQ_CMD_UNLOCK, nodepath)

    def getdescendentlocks(self, nodepath):
        """Return the decendent locks info.

        API command:
            /descendent/locks

        Args:
            nodepath (str): Path to a dirctory on the server.

        Returns:
            rpws.models.ChildrenLockInfo

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for folder in rserver.listfolders('/example/path'):
            ...     clockinfo = rserver.getdescendentlocks(folder.path)
            ...     for locked_model_path in clockinfo.items:
            ...         print(locked_model_path)
        """
        # get decendent lock data
        chlocks_dict = self._get(api.REQ_CMD_CHILDNLOCKS, nodepath)
        # get the locked children list
        chlocks = chlocks_dict[api.CHILDLOCKS_ITEMS_KEY]
        if chlocks:
            # corrent the paths so they're from root
            locked_childs = [self._root_path(x) for x in chlocks]
        else:
            locked_childs = []

        # make the children lock info obj
        return models.ChildrenLockInfo(
            path=nodepath,
            items=locked_childs,
            lock_context=chlocks_dict[api.CHILDLOCKS_LOCKCTX])

    def deletedescendentlocks(self, nodepath):
        """Delete the decendent locks.

        API command:
            /descendent/locks

        Args:
            nodepath (str): Path to a dirctory on the server.

        Returns:
            list of str for list of entries with failed locks

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for folder in rserver.listfolders('/example/path'):
            ...     rserver.deletedescendentlocks(folder.path)
        """
        # get decendent lock data
        chlocks_dict = self._delete(api.REQ_CMD_CHILDNLOCKS, nodepath)
        # get the list of failed locks
        failedchlocks = chlocks_dict[api.CHILDLOCKS_DELFAILEDITEMS_KEY]
        if failedchlocks:
            # corrent the paths so they're from root
            return [self._root_path(x) for x in failedchlocks]
        else:
            return []

    def mkdir(self, nodepath):
        """Create a new directory.

        Args:
            nodepath (str): Path to a dirctory on the server.

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> rserver.mkdir('/example/path')
        """
        return self._put(api.REQ_CMD_MKDIR, nodepath)

    def rename(self, nodepath, new_nodename):
        """Rename a file, folder, or model.

        Args:
            nodepath (str): Path to a file, folder, or model on the server.
            new_nodename (str): New name for file, folder, or model

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> rserver.rename('/example/path', '/example/newpath')
        """
        return self._delete(api.REQ_CMD_RENAME.format(new_name=new_nodename),
                            nodepath)

    def rmdir(self, nodepath):
        """Delete a file, folder, or model.

        Args:
            nodepath (str): Path to a file, folder, or model on the server.

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> rserver.rmdir('/example/path')
        """
        return self._delete(api.REQ_CMD_DELETE, nodepath)

    def delete(self, nodepath):
        """Delete a file, folder, or model.

        Args:
            nodepath (str): Path to a file, folder, or model on the server.

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> rserver.delete('/example/path/model.rvt')
        """
        return self._delete(api.REQ_CMD_DELETE, nodepath)

    def copy(self, nodepath, new_nodepath, overwrite=False):
        """Copy a file, folder, or model to new location.

        Args:
            nodepath (str): Path to a file, folder, or model on the server.
            new_nodepath (str): Full path to new location and name.
            overwrite (bool): True to overwrite any existing entries.

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> rserver.copy('/example/model.rvt', '/example/newmodel.rvt')
        """
        # get api path for the new location
        new_apipath = self._api_path(new_nodepath)
        return self._post(api.REQ_CMD_COPY.format(
            dest_path=new_apipath,
            replace_exist=overwrite), nodepath)

    def move(self, nodepath, new_nodepath, overwrite=False):
        """Move a file, folder, or model to new location.

        Args:
            nodepath (str): Path to a file, folder, or model on the server.
            new_nodepath (str): Full path to new location and name.
            overwrite (bool): True to overwrite any existing entries.

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> rserver.move('/example/model.rvt', '/example/newmodel.rvt')
        """
        # get api path for the new location
        new_apipath = self._api_path(new_nodepath)
        return self._post(api.REQ_CMD_MOVE.format(dest_path=new_apipath,
                                                  replace_exist=overwrite),
                          nodepath)

    def walk(self, top=None, topdown=True, digmodels=False):
        r"""Walk the provided path or root.

        Yields a 4-tuple of parent directory, folders, files, and models

        Args:
            top (str, optional): Parent directory. Root if not provided.
            topdown (bool): True to start from top and walk down
            digmodels (bool): True to list entries under a model folder
                              Revit models on Revit Server are actually folders
                              with files, and other subfolders.

        Returns:
            tuple: (parent, folders, files, models)

        Example:
            >>> rserver = RevitServer('server01', '2017')
            >>> for parent, folders, files, models in rserver.walk():
            ...     print(parent)
            ...     for fd in folders:
            ...         print('\t@d {}'.format(fd.path))
            ...     for f in files:
            ...         print('\t@f {}'.format(f.path))
            ...     for m in models:
            ...         print('\t@m {}'.format(m.path))
        """
        if not top:
            top = self.path

        entry_info = self.scandir(top)

        if topdown:
            # Yield before recursion if going top down
            yield top, entry_info.folders, entry_info.files, entry_info.models

        # Recurse into sub-directories
        for finfo in entry_info.folders:
            for x in self.walk(top=finfo.path, topdown=topdown):
                yield x

        # Recurse into sub-directories inside models
        if digmodels:
            for minfo in entry_info.models:
                for x in self.walk(top=minfo.path,
                                   topdown=topdown,
                                   digmodels=digmodels):
                    yield x
        if not topdown:
            # Yield before recursion if going top down
            yield top, entry_info.folders, entry_info.files, entry_info.models



class IDataObject:
 """ Provides a format-independent mechanism for transferring data. """
 def GetData(self,format,autoConvert=None):
  """
  GetData(self: IDataObject,format: Type) -> object

  

   Retrieves the data associated with the specified class type format.

  

   format: A System.Type representing the format of the data to retrieve. See 

    System.Windows.Forms.DataFormats for predefined formats.

  

   Returns: The data associated with the specified format,or null.

  GetData(self: IDataObject,format: str) -> object

  

   Retrieves the data associated with the specified data format.

  

   format: The format of the data to retrieve. See System.Windows.Forms.DataFormats for predefined formats.

   Returns: The data associated with the specified format,or null.

  GetData(self: IDataObject,format: str,autoConvert: bool) -> object

  

   Retrieves the data associated with the specified data format,using a Boolean to determine 

    whether to convert the data to the format.

  

  

   format: The format of the data to retrieve. See System.Windows.Forms.DataFormats for predefined formats.

   autoConvert: true to convert the data to the specified format; otherwise,false.

   Returns: The data associated with the specified format,or null.
  """
  pass
 def GetDataPresent(self,format,autoConvert=None):
  """
  GetDataPresent(self: IDataObject,format: Type) -> bool

  

   Determines whether data stored in this instance is associated with,or can be converted to,the 

    specified format.

  

  

   format: A System.Type representing the format for which to check. See System.Windows.Forms.DataFormats 

    for predefined formats.

  

   Returns: true if data stored in this instance is associated with,or can be converted to,the specified 

    format; otherwise,false.

  

  GetDataPresent(self: IDataObject,format: str) -> bool

  

   Determines whether data stored in this instance is associated with,or can be converted to,the 

    specified format.

  

  

   format: The format for which to check. See System.Windows.Forms.DataFormats for predefined formats.

   Returns: true if data stored in this instance is associated with,or can be converted to,the specified 

    format; otherwise false.

  

  GetDataPresent(self: IDataObject,format: str,autoConvert: bool) -> bool

  

   Determines whether data stored in this instance is associated with the specified format,using a 

    Boolean value to determine whether to convert the data to the format.

  

  

   format: The format for which to check. See System.Windows.Forms.DataFormats for predefined formats.

   autoConvert: true to determine whether data stored in this instance can be converted to the specified format; 

    false to check whether the data is in the specified format.

  

   Returns: true if the data is in,or can be converted to,the specified format; otherwise,false.
  """
  pass
 def GetFormats(self,autoConvert=None):
  """
  GetFormats(self: IDataObject) -> Array[str]

  

   Returns a list of all formats that data stored in this instance is associated with or can be 

    converted to.

  

   Returns: An array of the names that represents a list of all formats that are supported by the data 

    stored in this object.

  

  GetFormats(self: IDataObject,autoConvert: bool) -> Array[str]

  

   Gets a list of all formats that data stored in this instance is associated with or can be 

    converted to,using a Boolean value to determine whether to retrieve all formats that the data 

    can be converted to or only native data formats.

  

  

   autoConvert: true to retrieve all formats that data stored in this instance is associated with or can be 

    converted to; false to retrieve only native data formats.

  

   Returns: An array of the names that represents a list of all formats that are supported by the data 

    stored in this object.
  """
  pass
 def SetData(self,*__args):
  """
  SetData(self: IDataObject,format: Type,data: object)

   Stores the specified data and its associated class type in this instance.

  

   format: A System.Type representing the format associated with the data. See 

    System.Windows.Forms.DataFormats for predefined formats.

  

   data: The data to store.

  SetData(self: IDataObject,data: object)

   Stores the specified data in this instance,using the class of the data for the format.

  

   data: The data to store.

  SetData(self: IDataObject,format: str,autoConvert: bool,data: object)

   Stores the specified data and its associated format in this instance,using a Boolean value to 

    specify whether the data can be converted to another format.

  

  

   format: The format associated with the data. See System.Windows.Forms.DataFormats for predefined formats.

   autoConvert: true to allow the data to be converted to another format; otherwise,false.

   data: The data to store.

  SetData(self: IDataObject,format: str,data: object)

   Stores the specified data and its associated format in this instance.

  

   format: The format associated with the data. See System.Windows.Forms.DataFormats for predefined formats.

   data: The data to store.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class ColumnHeader(Component,IComponent,IDisposable,ICloneable):
 """
 Displays a single column header in a System.Windows.Forms.ListView control.

 

 ColumnHeader()

 ColumnHeader(imageIndex: int)

 ColumnHeader(imageKey: str)
 """
 def AutoResize(self,headerAutoResize):
  """
  AutoResize(self: ColumnHeader,headerAutoResize: ColumnHeaderAutoResizeStyle)

   Resizes the width of the column as indicated by the resize style.

  

   headerAutoResize: One of the System.Windows.Forms.ColumnHeaderAutoResizeStyle  values.
  """
  pass
 def Clone(self):
  """
  Clone(self: ColumnHeader) -> object

  

   Creates an identical copy of the current System.Windows.Forms.ColumnHeader that is not attached 

    to any list view control.

  

   Returns: An object representing a copy of this System.Windows.Forms.ColumnHeader object.
  """
  pass
 def Dispose(self):
  """
  Dispose(self: ColumnHeader,disposing: bool)

   Disposes of the resources (other than memory) used by the System.Windows.Forms.ColumnHeader.

  

   disposing: true to release both managed and unmanaged resources; false to release only unmanaged resources.
  """
  pass
 def GetService(self,*args):
  """
  GetService(self: Component,service: Type) -> object

  

   Returns an object that represents a service provided by the System.ComponentModel.Component or 

    by its System.ComponentModel.Container.

  

  

   service: A service provided by the System.ComponentModel.Component.

   Returns: An System.Object that represents a service provided by the System.ComponentModel.Component,or 

    null if the System.ComponentModel.Component does not provide the specified service.
  """
  pass
 def MemberwiseClone(self,*args):
  """
  MemberwiseClone(self: MarshalByRefObject,cloneIdentity: bool) -> MarshalByRefObject

  

   Creates a shallow copy of the current System.MarshalByRefObject object.

  

   cloneIdentity: false to delete the current System.MarshalByRefObject object's identity,which will cause the 

    object to be assigned a new identity when it is marshaled across a remoting boundary. A value of 

    false is usually appropriate. true to copy the current System.MarshalByRefObject object's 

    identity to its clone,which will cause remoting client calls to be routed to the remote server 

    object.

  

   Returns: A shallow copy of the current System.MarshalByRefObject object.

  MemberwiseClone(self: object) -> object

  

   Creates a shallow copy of the current System.Object.

   Returns: A shallow copy of the current System.Object.
  """
  pass
 def ToString(self):
  """
  ToString(self: ColumnHeader) -> str

  

   Returns a string representation of this column header.

   Returns: A System.String containing the name of the System.ComponentModel.Component,if any,or null if 

    the System.ComponentModel.Component is unnamed.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type)

  __new__(cls: type,imageIndex: int)

  __new__(cls: type,imageKey: str)
  """
  pass
 def __str__(self,*args):
  pass
 CanRaiseEvents=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the component can raise an event.



"""

 DesignMode=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the System.ComponentModel.Component is currently in design mode.



"""

 DisplayIndex=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the display order of the column relative to the currently displayed columns.



Get: DisplayIndex(self: ColumnHeader) -> int



Set: DisplayIndex(self: ColumnHeader)=value

"""

 Events=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the list of event handlers that are attached to this System.ComponentModel.Component.



"""

 ImageIndex=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the index of the image displayed in the System.Windows.Forms.ColumnHeader.



Get: ImageIndex(self: ColumnHeader) -> int



Set: ImageIndex(self: ColumnHeader)=value

"""

 ImageKey=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the key of the image displayed in the column.



Get: ImageKey(self: ColumnHeader) -> str



Set: ImageKey(self: ColumnHeader)=value

"""

 ImageList=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the image list associated with the System.Windows.Forms.ColumnHeader.



Get: ImageList(self: ColumnHeader) -> ImageList



"""

 Index=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the location with the System.Windows.Forms.ListView control's System.Windows.Forms.ListView.ColumnHeaderCollection of this column.



Get: Index(self: ColumnHeader) -> int



"""

 ListView=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Windows.Forms.ListView control the System.Windows.Forms.ColumnHeader is located in.



Get: ListView(self: ColumnHeader) -> ListView



"""

 Name=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the name of the System.Windows.Forms.ColumnHeader.



Get: Name(self: ColumnHeader) -> str



Set: Name(self: ColumnHeader)=value

"""

 Tag=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets an object that contains data to associate with the System.Windows.Forms.ColumnHeader.



Get: Tag(self: ColumnHeader) -> object



Set: Tag(self: ColumnHeader)=value

"""

 Text=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the text displayed in the column header.



Get: Text(self: ColumnHeader) -> str



Set: Text(self: ColumnHeader)=value

"""

 TextAlign=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the horizontal alignment of the text displayed in the System.Windows.Forms.ColumnHeader.



Get: TextAlign(self: ColumnHeader) -> HorizontalAlignment



Set: TextAlign(self: ColumnHeader)=value

"""

 Width=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the width of the column.



Get: Width(self: ColumnHeader) -> int



Set: Width(self: ColumnHeader)=value

"""





class DialogResult(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies identifiers to indicate the return value of a dialog box.

 

 enum DialogResult,values: Abort (3),Cancel (2),Ignore (5),No (7),None (0),OK (1),Retry (4),Yes (6)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Abort=None
 Cancel=None
 Ignore=None
 No=None
 None=None
 OK=None
 Retry=None
 value__=None
 Yes=None




class CookieContainer(object):
 """
 Provides a container for a collection of System.Net.CookieCollection objects.

 

 CookieContainer()

 CookieContainer(capacity: int)

 CookieContainer(capacity: int,perDomainCapacity: int,maxCookieSize: int)
 """
 def Add(self,*__args):
  """
  Add(self: CookieContainer,uri: Uri,cookie: Cookie)

   Adds a System.Net.Cookie to the System.Net.CookieContainer for a particular URI.

  

   uri: The URI of the System.Net.Cookie to be added to the System.Net.CookieContainer.

   cookie: The System.Net.Cookie to be added to the System.Net.CookieContainer.

  Add(self: CookieContainer,uri: Uri,cookies: CookieCollection)

   Adds the contents of a System.Net.CookieCollection to the System.Net.CookieContainer for a 

    particular URI.

  

  

   uri: The URI of the System.Net.CookieCollection to be added to the System.Net.CookieContainer.

   cookies: The System.Net.CookieCollection to be added to the System.Net.CookieContainer.

  Add(self: CookieContainer,cookie: Cookie)

   Adds a System.Net.Cookie to a System.Net.CookieContainer. This method uses the domain from the 

    System.Net.Cookie to determine which domain collection to associate the System.Net.Cookie with.

  

  

   cookie: The System.Net.Cookie to be added to the System.Net.CookieContainer.

  Add(self: CookieContainer,cookies: CookieCollection)

   Adds the contents of a System.Net.CookieCollection to the System.Net.CookieContainer.

  

   cookies: The System.Net.CookieCollection to be added to the System.Net.CookieContainer.
  """
  pass
 def GetCookieHeader(self,uri):
  """
  GetCookieHeader(self: CookieContainer,uri: Uri) -> str

  

   Gets the HTTP cookie header that contains the HTTP cookies that represent the System.Net.Cookie 

    instances that are associated with a specific URI.

  

  

   uri: The URI of the System.Net.Cookie instances desired.

   Returns: An HTTP cookie header,with strings representing System.Net.Cookie instances delimited by 

    semicolons.
  """
  pass
 def GetCookies(self,uri):
  """
  GetCookies(self: CookieContainer,uri: Uri) -> CookieCollection

  

   Gets a System.Net.CookieCollection that contains the System.Net.Cookie instances that are 

    associated with a specific URI.

  

  

   uri: The URI of the System.Net.Cookie instances desired.

   Returns: A System.Net.CookieCollection that contains the System.Net.Cookie instances that are associated 

    with a specific URI.
  """
  pass
 def SetCookies(self,uri,cookieHeader):
  """
  SetCookies(self: CookieContainer,uri: Uri,cookieHeader: str)

   Adds System.Net.Cookie instances for one or more cookies from an HTTP cookie header to the 

    System.Net.CookieContainer for a specific URI.

  

  

   uri: The URI of the System.Net.CookieCollection.

   cookieHeader: The contents of an HTTP set-cookie header as returned by a HTTP server,with System.Net.Cookie 

    instances delimited by commas.
  """
  pass
 def __add__(self,*args):
  """ x.__add__(y) <==> x+yx.__add__(y) <==> x+yx.__add__(y) <==> x+yx.__add__(y) <==> x+y """
  pass
 @staticmethod
 def __new__(self,capacity=None,perDomainCapacity=None,maxCookieSize=None):
  """
  __new__(cls: type)

  __new__(cls: type,capacity: int)

  __new__(cls: type,capacity: int,perDomainCapacity: int,maxCookieSize: int)
  """
  pass
 Capacity=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets and sets the number of System.Net.Cookie instances that a System.Net.CookieContainer can hold.



Get: Capacity(self: CookieContainer) -> int



Set: Capacity(self: CookieContainer)=value

"""

 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of System.Net.Cookie instances that a System.Net.CookieContainer currently holds.



Get: Count(self: CookieContainer) -> int



"""

 MaxCookieSize=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Represents the maximum allowed length of a System.Net.Cookie.



Get: MaxCookieSize(self: CookieContainer) -> int



Set: MaxCookieSize(self: CookieContainer)=value

"""

 PerDomainCapacity=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets and sets the number of System.Net.Cookie instances that a System.Net.CookieContainer can hold per domain.



Get: PerDomainCapacity(self: CookieContainer) -> int



Set: PerDomainCapacity(self: CookieContainer)=value

"""


 DefaultCookieLengthLimit=4096
 DefaultCookieLimit=300
 DefaultPerDomainCookieLimit=20




class HtmlHistory(object,IDisposable):
 """ Manages the list of documents and Web sites the user has visited within the current session. """
 def Back(self,numberBack):
  """
  Back(self: HtmlHistory,numberBack: int)

   Navigates backward in the navigation stack by the specified number of entries.

  

   numberBack: The number of entries to navigate backward in the navigation stack. This number must be a 

    positive integer.
  """
  pass
 def Dispose(self):
  """
  Dispose(self: HtmlHistory)

   Releases all resources used by the System.Windows.Forms.HtmlHistory.
  """
  pass
 def Forward(self,numberForward):
  """
  Forward(self: HtmlHistory,numberForward: int)

   Navigates forward in the navigation stack by the specified number of entries.

  

   numberForward: The number of entries to navigate forward in the navigation stack. This number must be a 

    positive integer.
  """
  pass
 def Go(self,*__args):
  """
  Go(self: HtmlHistory,relativePosition: int)

   Navigates to the specified relative position in the browser's history.

  

   relativePosition: The entry in the navigation stack you want to display.

  Go(self: HtmlHistory,urlString: str)

   Navigates to the specified Uniform Resource Locator (URL).

  

   urlString: The URL you want to display. This may be a relative or virtual URL (for example,page.html,

    path/page.html,or /path/to/page.html),in which case the current Web page's URL is used as a 

    base.

  

  Go(self: HtmlHistory,url: Uri)

   Navigates to the specified Uniform Resource Locator (URL).

  

   url: The URL as a System.Uri object.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 DomHistory=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the unmanaged interface wrapped by this class.



Get: DomHistory(self: HtmlHistory) -> object



"""

 Length=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the size of the history stack.



Get: Length(self: HtmlHistory) -> int



"""





class DataGridCellEditEndingEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Controls.DataGrid.CellEditEnding event.

 

 DataGridCellEditEndingEventArgs(column: DataGridColumn,row: DataGridRow,editingElement: FrameworkElement,editAction: DataGridEditAction)
 """
 @staticmethod
 def __new__(self,column,row,editingElement,editAction):
  """ __new__(cls: type,column: DataGridColumn,row: DataGridRow,editingElement: FrameworkElement,editAction: DataGridEditAction) """
  pass
 Cancel=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value that indicates whether the event should be canceled.



Get: Cancel(self: DataGridCellEditEndingEventArgs) -> bool



Set: Cancel(self: DataGridCellEditEndingEventArgs)=value

"""

 Column=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the column that contains the cell for which the event occurred.



Get: Column(self: DataGridCellEditEndingEventArgs) -> DataGridColumn



"""

 EditAction=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the edit was canceled or committed.



Get: EditAction(self: DataGridCellEditEndingEventArgs) -> DataGridEditAction



"""

 EditingElement=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the element that the cell displays in editing mode.



Get: EditingElement(self: DataGridCellEditEndingEventArgs) -> FrameworkElement



"""

 Row=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the row that contains the cell for which the event occurred.



Get: Row(self: DataGridCellEditEndingEventArgs) -> DataGridRow



"""


